# Copyright 2017 Dimitri Capitaine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from ._vendor.future.standard_library import install_aliases
install_aliases()

import urllib.parse
import lxdbapi

from sqlalchemy.engine.default import DefaultDialect

from sqlalchemy.sql.compiler import DDLCompiler, SQLCompiler
from sqlalchemy.sql import compiler
from sqlalchemy.engine import default

from sqlalchemy import types, BigInteger, Float
from sqlalchemy.types import INTEGER, BIGINT, SMALLINT, VARCHAR, CHAR, \
    FLOAT, DATE, BOOLEAN, DECIMAL, TIMESTAMP, TIME, VARBINARY


class LeanxcaleCompiler(SQLCompiler):
    def visit_sequence(self, seq):
        return "NEXT VALUE FOR %s" % seq.name

class LeanxcaleDDLCompiler(DDLCompiler):

    def visit_primary_key_constraint(self, constraint):
        # if constraint.name is None:
        #    raise CompileError("can't create primary key without a name")
        return DDLCompiler.visit_primary_key_constraint(self, constraint)

    def visit_sequence(self, seq):
        return "NEXT VALUE FOR %s" % seq.name

class LeanxcaleExecutionContext(default.DefaultExecutionContext):

    def should_autocommit_text(self, statement):
        pass

    def create_server_side_cursor(self):
        pass

    def fire_sequence(self, seq, type_):
        return self._execute_scalar(
            (
                    "SELECT NEXT VALUE FOR %s"
                    % seq.name
            ),
            type_
        )


class LeanXcaleIdentifierPreparer(compiler.IdentifierPreparer):

    def __init__(self, dialect, server_ansiquotes=False, **kw):

        quote = '"'

        super(LeanXcaleIdentifierPreparer, self).__init__(
            dialect, initial_quote=quote, escape_quote=quote
        )

    def _quote_free_identifiers(self, *ids):
        """Unilaterally identifier-quote any number of strings."""

        return tuple([self.quote_identifier(i) for i in ids if i is not None])

class LeanXcaleGenericTypeCompiler(compiler.GenericTypeCompiler):
    def visit_ARRAY(self, type_, **kw):
        if type_.item_type.python_type == int:
            return "BIGINT ARRAY"
        elif type_.item_type.python_type == float:
            return "DOUBLE ARRAY"
        elif type_.item_type.python_type == str:
            return "VARCHAR ARRAY"
        else:
            raise Exception("ARRAY of type {} is not supported".format(str(type_)))


class LeanxcaleDialect(DefaultDialect):
    name = "leanxcale"

    driver = "lxdbapi"

    ddl_compiler = LeanxcaleDDLCompiler
    preparer = LeanXcaleIdentifierPreparer
    statement_compiler = LeanxcaleCompiler
    type_compiler = LeanXcaleGenericTypeCompiler

    supports_sequences = True
    sequences_optional = True
    supports_multivalues_insert = True

    execution_ctx_cls = LeanxcaleExecutionContext

    preexecute_autoincrement_sequences = True

    @classmethod
    def dbapi(cls):

        return lxdbapi

    def create_connect_args(self, url):

        # Deal with PARALLEL option
        parallel = url.query.get('parallel')

        # Deal with SECURE option
        secure = url.query.get('secure')

        distribute = url.query.get('distribute')

        leanxcale_url = urllib.parse.urlunsplit(urllib.parse.SplitResult(
            scheme='http',
            netloc='{}:{}'.format(url.host, url.port or 8765),
            path='/',
            query=urllib.parse.urlencode(url.query),
            fragment='',
        ))
        if (url.query.get('autocommit') == 'False'):
            autocommit = False
        else:
            autocommit = True
        # By default, do not use timezoned timestamps
        if (url.query.get('timezonedTimestamps') == 'True'):
            timezone = True
        else:
            timezone = False

        params = {'autocommit': autocommit, 'user': url.username}
        if parallel is not None:
            params.update({'parallel': parallel})
        if secure is not None:
            params.update({'secure': secure})
        if distribute is not None:
            params.update({'distribute': distribute})
        if url.password is not None:
            params.update({'password': url.password})
        params.update({'useTZ': timezone})

        params.update({'databaseName': url.database})
        return [leanxcale_url], params

    def _get_default_schema_name(self, connection):
        return connection.engine.url.username.upper()

    def do_rollback(self, dbapi_conection):
        dbapi_conection.rollback()

    def do_commit(self, dbapi_conection):
        dbapi_conection.commit()

    def has_sequence(self, connection, sequence_name, schema=None, **kw):
        return sequence_name.upper() in self.get_sequence_names(connection, schema)

    def get_sequence_names(self, connection, schema=None, **kw):
        dbapi_con = connection.connect().connection
        cursor = dbapi_con.cursor()
        typeList = ["SEQUENCE"]
        schema = schema.upper() if schema is not None else self.default_schema_name
        cursor.get_tables(schemaPattern=schema, typeList=typeList)
        sequences = cursor.fetchall()

        result = []
        for seq in sequences:
            result.append(seq[2])
        return result

    def has_table(self, connection, table_name, schema=None, **kw):
        return table_name.upper() in self.get_table_names(connection, schema)

    def get_schema_names(self, connection, catalog=None, schema=None, **kw):
        dbapi_con = connection.connect().connection
        cursor = dbapi_con.cursor()
        schema = schema.upper() if schema is not None else self.default_schema_name
        cursor.get_schemas(catalog, schema)
        schemas = cursor.fetchall()

        result = []
        for schema in schemas:
            result.append(schema[0])
        return result

    def get_table_names(self, connection, schema=None, **kw):

        dbapi_con = connection.connect().connection
        cursor = dbapi_con.cursor()
        tableNamePattern = None if 'tableNamePattern' not in kw.keys() else kw.get('tableNamePattern')
        typeList = ['TABLE', 'MATERIALIZED QUERY TABLE'] if 'typeList' not in kw.keys() else kw.get('typeList')
        schema = schema.upper() if schema is not None else self.default_schema_name
        cursor.get_tables(schemaPattern=schema, tableNamePattern=tableNamePattern, typeList=typeList)
        tables = cursor.fetchall()
        result = []
        if tables:
            for table in tables:
                result.append(table[2])
        return result

    def get_columns(self, connection, table_name=None, schema=None, catalog=None, column_name=None, **kw):

        dbapi_con = connection.connect().connection
        cursor = dbapi_con.cursor()
        schema = schema.upper() if schema is not None else self.default_schema_name
        cursor.get_columns(catalog, schema, table_name, column_name)
        columns = cursor.fetchall()

        result = []
        for column in columns:
            col_d = {}
            col_d.update({'name': column[3]})
            if column[4] == 2003:
                if column[5].split()[0] == 'BIGINT':
                    col_d.update({'item_type': BigInteger})
                elif column[5].split()[0] == 'DOUBLE':
                    col_d.update({'item_type': Float})
                elif column[5].split()[0] == 'VARCHAR':
                    col_d.update({'item_type': VARCHAR})
                col_d.update({'type': ARRAY(column[5])})
            else:
                col_d.update({'type': COLUMN_DATA_TYPE[column[4]]})
            col_d.update({'nullable': column[10] == 1 if True else False})
            col_d.update({'default': column[13]})

            result.append(col_d)

        return result

    def get_view_names(self, connection, schema=None, **kw):
        dbapi_con = connection.connect().connection
        cursor = dbapi_con.cursor()
        schema = schema.upper() if schema is not None else self.default_schema_name
        cursor.get_tables(schemaPattern=schema, typeList=['VIEW'])
        tables = cursor.fetchall()

        result = []
        for table in tables:
            result.append(table[2])
        return result

    def get_pk_constraint(self, connection, table_name, schema=None, **kw):
        pk_sql = "SELECT * FROM LXSYSMETA.PRIMARY_KEYS " \
                 "WHERE TABLENAME = '" + table_name +"' ORDER BY keyseq"
        dbapi_con = connection.connect()
        rs = dbapi_con.execute(pk_sql)
        pk_list = [r for r in rs]
        print("pk_list: {}".format(pk_list))
        result = {
            'constrained_columns': [],
        }
        [result['constrained_columns'].append(pk[0]) for pk in pk_list]
        return result

    def get_foreign_keys(self, connection, table_name, schema=None, **kw):
        return []

    def get_indexes(self, connection, table_name, schema=None, **kw):
        index_sql = "SELECT * FROM LXSYSMETA.INDEXES WHERE TABLENAME = '" + table_name +"' order by indexname, ordinalposition"
        dbapi_con = connection.connect()
        rs = dbapi_con.execute(index_sql)
        index_list = [r for r in rs]

        result = []

        for index in index_list:
            name = index[2]
            column = index[5]
            nonUnique = index[0]

            if len(result) == 0:
                result.append({
                    'name': name,
                    'column_names': [column],
                    'unique': not nonUnique
                })
            else:
                found = False
                for final_index in result:
                    if name == final_index.get('name'):
                        found = True
                        final_index['column_names'].append(column)
                if not found:
                    result.append({
                        'name': name,
                        'column_names': [column],
                        'unique': not nonUnique
                    })
        return result

    #def has_index(self, connection, table_name, index_name, schema=None):
    #    pass

class TINYINT(types.Integer):
    __visit_name__ = "INTEGER"


class UTINYINT(types.Integer):
    __visit_name__ = "INTEGER"


class UINTEGER(types.Integer):
    __visit_name__ = "INTEGER"


class DOUBLE(types.BIGINT):
    __visit_name__ = "BIGINT"


class DOUBLE(types.BIGINT):
    __visit_name__ = "BIGINT"


class UDOUBLE(types.BIGINT):
    __visit_name__ = "BIGINT"


class UFLOAT(types.FLOAT):
    __visit_name__ = "FLOAT"


class ULONG(types.BIGINT):
    __visit_name__ = "BIGINT"


class UTIME(types.TIME):
    __visit_name__ = "TIME"


class UDATE(types.DATE):
    __visit_name__ = "DATE"


class UTIMESTAMP(types.TIMESTAMP):
    __visit_name__ = "TIMESTAMP"


class ROWID(types.String):
    __visit_name__ = "VARCHAR"

class ARRAY(types.ARRAY):
    __visit_name__ = "ARRAY"
    def __init__(self, type):
        types.ARRAY.__init__(self, type)



COLUMN_DATA_TYPE = {
    -6: TINYINT,
    -5: BIGINT,
    -3: VARBINARY,
    1: CHAR,
    3: DECIMAL,
    4: INTEGER,
    5: SMALLINT,
    6: FLOAT,
    8: DOUBLE,
    9: UINTEGER,
    10: ULONG,
    11: UTINYINT,
    12: VARCHAR,
    13: ROWID,
    14: UFLOAT,
    15: UDOUBLE,
    16: BOOLEAN,
    18: UTIME,
    19: UDATE,
    20: UTIMESTAMP,
    91: DATE,
    92: TIME,
    93: TIMESTAMP
}



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: lxdbapi/common.proto
"""Generated protocol buffer code."""
from ._vendor.google.protobuf.internal import enum_type_wrapper
from ._vendor.google.protobuf import descriptor as _descriptor
from ._vendor.google.protobuf import descriptor_pool as _descriptor_pool
from ._vendor.google.protobuf import message as _message
from ._vendor.google.protobuf import reflection as _reflection
from ._vendor.google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14lxdbapi/common.proto\x12\x07lxdbapi\"\xc0\x01\n\x14\x43onnectionProperties\x12\x10\n\x08is_dirty\x18\x01 \x01(\x08\x12\x13\n\x0b\x61uto_commit\x18\x02 \x01(\x08\x12\x17\n\x0fhas_auto_commit\x18\x07 \x01(\x08\x12\x11\n\tread_only\x18\x03 \x01(\x08\x12\x15\n\rhas_read_only\x18\x08 \x01(\x08\x12\x1d\n\x15transaction_isolation\x18\x04 \x01(\r\x12\x0f\n\x07\x63\x61talog\x18\x05 \x01(\t\x12\x0e\n\x06schema\x18\x06 \x01(\t\"[\n\x0fStatementHandle\x12\x15\n\rconnection_id\x18\x01 \x01(\t\x12\n\n\x02id\x18\x02 \x01(\r\x12%\n\tsignature\x18\x03 \x01(\x0b\x32\x12.lxdbapi.Signature\"\xd0\x01\n\tSignature\x12(\n\x07\x63olumns\x18\x01 \x03(\x0b\x32\x17.lxdbapi.ColumnMetaData\x12\x0b\n\x03sql\x18\x02 \x01(\t\x12-\n\nparameters\x18\x03 \x03(\x0b\x32\x19.lxdbapi.AvaticaParameter\x12.\n\x0e\x63ursor_factory\x18\x04 \x01(\x0b\x32\x16.lxdbapi.CursorFactory\x12-\n\rstatementType\x18\x05 \x01(\x0e\x32\x16.lxdbapi.StatementType\"\xb5\x03\n\x0e\x43olumnMetaData\x12\x0f\n\x07ordinal\x18\x01 \x01(\r\x12\x16\n\x0e\x61uto_increment\x18\x02 \x01(\x08\x12\x16\n\x0e\x63\x61se_sensitive\x18\x03 \x01(\x08\x12\x12\n\nsearchable\x18\x04 \x01(\x08\x12\x10\n\x08\x63urrency\x18\x05 \x01(\x08\x12\x10\n\x08nullable\x18\x06 \x01(\r\x12\x0e\n\x06signed\x18\x07 \x01(\x08\x12\x14\n\x0c\x64isplay_size\x18\x08 \x01(\r\x12\r\n\x05label\x18\t \x01(\t\x12\x13\n\x0b\x63olumn_name\x18\n \x01(\t\x12\x13\n\x0bschema_name\x18\x0b \x01(\t\x12\x11\n\tprecision\x18\x0c \x01(\r\x12\r\n\x05scale\x18\r \x01(\r\x12\x12\n\ntable_name\x18\x0e \x01(\t\x12\x14\n\x0c\x63\x61talog_name\x18\x0f \x01(\t\x12\x11\n\tread_only\x18\x10 \x01(\x08\x12\x10\n\x08writable\x18\x11 \x01(\x08\x12\x1b\n\x13\x64\x65\x66initely_writable\x18\x12 \x01(\x08\x12\x19\n\x11\x63olumn_class_name\x18\x13 \x01(\t\x12\"\n\x04type\x18\x14 \x01(\x0b\x32\x14.lxdbapi.AvaticaType\"\x95\x01\n\x0b\x41vaticaType\x12\n\n\x02id\x18\x01 \x01(\r\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x19\n\x03rep\x18\x03 \x01(\x0e\x32\x0c.lxdbapi.Rep\x12(\n\x07\x63olumns\x18\x04 \x03(\x0b\x32\x17.lxdbapi.ColumnMetaData\x12\'\n\tcomponent\x18\x05 \x01(\x0b\x32\x14.lxdbapi.AvaticaType\"\x91\x01\n\x10\x41vaticaParameter\x12\x0e\n\x06signed\x18\x01 \x01(\x08\x12\x11\n\tprecision\x18\x02 \x01(\r\x12\r\n\x05scale\x18\x03 \x01(\r\x12\x16\n\x0eparameter_type\x18\x04 \x01(\r\x12\x11\n\ttype_name\x18\x05 \x01(\t\x12\x12\n\nclass_name\x18\x06 \x01(\t\x12\x0c\n\x04name\x18\x07 \x01(\t\"\xbb\x01\n\rCursorFactory\x12+\n\x05style\x18\x01 \x01(\x0e\x32\x1c.lxdbapi.CursorFactory.Style\x12\x12\n\nclass_name\x18\x02 \x01(\t\x12\x13\n\x0b\x66ield_names\x18\x03 \x03(\t\"T\n\x05Style\x12\n\n\x06OBJECT\x10\x00\x12\n\n\x06RECORD\x10\x01\x12\x15\n\x11RECORD_PROJECTION\x10\x02\x12\t\n\x05\x41RRAY\x10\x03\x12\x08\n\x04LIST\x10\x04\x12\x07\n\x03MAP\x10\x05\"A\n\x05\x46rame\x12\x0e\n\x06offset\x18\x01 \x01(\x04\x12\x0c\n\x04\x64one\x18\x02 \x01(\x08\x12\x1a\n\x04rows\x18\x03 \x03(\x0b\x32\x0c.lxdbapi.Row\"*\n\x03Row\x12#\n\x05value\x18\x01 \x03(\x0b\x32\x14.lxdbapi.ColumnValue\"3\n\x10\x44\x61tabaseProperty\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x11\n\tfunctions\x18\x02 \x03(\t\"4\n\x0bWireMessage\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x17\n\x0fwrapped_message\x18\x02 \x01(\x0c\"\x9f\x01\n\x0b\x43olumnValue\x12\"\n\x05value\x18\x01 \x03(\x0b\x32\x13.lxdbapi.TypedValue\x12(\n\x0b\x61rray_value\x18\x02 \x03(\x0b\x32\x13.lxdbapi.TypedValue\x12\x17\n\x0fhas_array_value\x18\x03 \x01(\x08\x12)\n\x0cscalar_value\x18\x04 \x01(\x0b\x32\x13.lxdbapi.TypedValue\"\x8a\x02\n\nTypedValue\x12\x1a\n\x04type\x18\x01 \x01(\x0e\x32\x0c.lxdbapi.Rep\x12\x12\n\nbool_value\x18\x02 \x01(\x08\x12\x14\n\x0cstring_value\x18\x03 \x01(\t\x12\x14\n\x0cnumber_value\x18\x04 \x01(\x12\x12\x13\n\x0b\x62ytes_value\x18\x05 \x01(\x0c\x12\x14\n\x0c\x64ouble_value\x18\x06 \x01(\x01\x12\x0c\n\x04null\x18\x07 \x01(\x08\x12(\n\x0b\x61rray_value\x18\x08 \x03(\x0b\x32\x13.lxdbapi.TypedValue\x12$\n\x0e\x63omponent_type\x18\t \x01(\x0e\x32\x0c.lxdbapi.Rep\x12\x17\n\x0fimplicitly_null\x18\n \x01(\x08\"\xae\x02\n\x19MetaDataOperationArgument\x12\x14\n\x0cstring_value\x18\x01 \x01(\t\x12\x12\n\nbool_value\x18\x02 \x01(\x08\x12\x11\n\tint_value\x18\x03 \x01(\x11\x12\x1b\n\x13string_array_values\x18\x04 \x03(\t\x12\x18\n\x10int_array_values\x18\x05 \x03(\x11\x12=\n\x04type\x18\x06 \x01(\x0e\x32/.lxdbapi.MetaDataOperationArgument.ArgumentType\"^\n\x0c\x41rgumentType\x12\n\n\x06STRING\x10\x00\x12\x08\n\x04\x42OOL\x10\x01\x12\x07\n\x03INT\x10\x02\x12\x13\n\x0fREPEATED_STRING\x10\x03\x12\x10\n\x0cREPEATED_INT\x10\x04\x12\x08\n\x04NULL\x10\x05\"\xc8\x01\n\nQueryState\x12 \n\x04type\x18\x01 \x01(\x0e\x32\x12.lxdbapi.StateType\x12\x0b\n\x03sql\x18\x02 \x01(\t\x12&\n\x02op\x18\x03 \x01(\x0e\x32\x1a.lxdbapi.MetaDataOperation\x12\x30\n\x04\x61rgs\x18\x04 \x03(\x0b\x32\".lxdbapi.MetaDataOperationArgument\x12\x10\n\x08has_args\x18\x05 \x01(\x08\x12\x0f\n\x07has_sql\x18\x06 \x01(\x08\x12\x0e\n\x06has_op\x18\x07 \x01(\x08\"?\n\x0cKeyValuePair\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\"\n\x05value\x18\x02 \x01(\x0b\x32\x13.lxdbapi.TypedValue*\x9f\x01\n\rStatementType\x12\n\n\x06SELECT\x10\x00\x12\n\n\x06INSERT\x10\x01\x12\n\n\x06UPDATE\x10\x02\x12\n\n\x06\x44\x45LETE\x10\x03\x12\n\n\x06UPSERT\x10\x04\x12\t\n\x05MERGE\x10\x05\x12\r\n\tOTHER_DML\x10\x06\x12\n\n\x06\x43REATE\x10\x07\x12\x08\n\x04\x44ROP\x10\x08\x12\t\n\x05\x41LTER\x10\t\x12\r\n\tOTHER_DDL\x10\n\x12\x08\n\x04\x43\x41LL\x10\x0b*\xe2\x03\n\x03Rep\x12\x15\n\x11PRIMITIVE_BOOLEAN\x10\x00\x12\x12\n\x0ePRIMITIVE_BYTE\x10\x01\x12\x12\n\x0ePRIMITIVE_CHAR\x10\x02\x12\x13\n\x0fPRIMITIVE_SHORT\x10\x03\x12\x11\n\rPRIMITIVE_INT\x10\x04\x12\x12\n\x0ePRIMITIVE_LONG\x10\x05\x12\x13\n\x0fPRIMITIVE_FLOAT\x10\x06\x12\x14\n\x10PRIMITIVE_DOUBLE\x10\x07\x12\x0b\n\x07\x42OOLEAN\x10\x08\x12\x08\n\x04\x42YTE\x10\t\x12\r\n\tCHARACTER\x10\n\x12\t\n\x05SHORT\x10\x0b\x12\x0b\n\x07INTEGER\x10\x0c\x12\x08\n\x04LONG\x10\r\x12\t\n\x05\x46LOAT\x10\x0e\x12\n\n\x06\x44OUBLE\x10\x0f\x12\x0f\n\x0b\x42IG_INTEGER\x10\x19\x12\x0f\n\x0b\x42IG_DECIMAL\x10\x1a\x12\x11\n\rJAVA_SQL_TIME\x10\x10\x12\x16\n\x12JAVA_SQL_TIMESTAMP\x10\x11\x12\x11\n\rJAVA_SQL_DATE\x10\x12\x12\x12\n\x0eJAVA_UTIL_DATE\x10\x13\x12\x0f\n\x0b\x42YTE_STRING\x10\x14\x12\n\n\x06STRING\x10\x15\x12\n\n\x06NUMBER\x10\x16\x12\n\n\x06OBJECT\x10\x17\x12\x08\n\x04NULL\x10\x18\x12\t\n\x05\x41RRAY\x10\x1b\x12\n\n\x06STRUCT\x10\x1c\x12\x0c\n\x08MULTISET\x10\x1d*^\n\x08Severity\x12\x14\n\x10UNKNOWN_SEVERITY\x10\x00\x12\x12\n\x0e\x46\x41TAL_SEVERITY\x10\x01\x12\x12\n\x0e\x45RROR_SEVERITY\x10\x02\x12\x14\n\x10WARNING_SEVERITY\x10\x03*\xd7\x04\n\x11MetaDataOperation\x12\x12\n\x0eGET_ATTRIBUTES\x10\x00\x12\x1b\n\x17GET_BEST_ROW_IDENTIFIER\x10\x01\x12\x10\n\x0cGET_CATALOGS\x10\x02\x12\x1e\n\x1aGET_CLIENT_INFO_PROPERTIES\x10\x03\x12\x19\n\x15GET_COLUMN_PRIVILEGES\x10\x04\x12\x0f\n\x0bGET_COLUMNS\x10\x05\x12\x17\n\x13GET_CROSS_REFERENCE\x10\x06\x12\x15\n\x11GET_EXPORTED_KEYS\x10\x07\x12\x18\n\x14GET_FUNCTION_COLUMNS\x10\x08\x12\x11\n\rGET_FUNCTIONS\x10\t\x12\x15\n\x11GET_IMPORTED_KEYS\x10\n\x12\x12\n\x0eGET_INDEX_INFO\x10\x0b\x12\x14\n\x10GET_PRIMARY_KEYS\x10\x0c\x12\x19\n\x15GET_PROCEDURE_COLUMNS\x10\r\x12\x12\n\x0eGET_PROCEDURES\x10\x0e\x12\x16\n\x12GET_PSEUDO_COLUMNS\x10\x0f\x12\x0f\n\x0bGET_SCHEMAS\x10\x10\x12\x19\n\x15GET_SCHEMAS_WITH_ARGS\x10\x11\x12\x14\n\x10GET_SUPER_TABLES\x10\x12\x12\x13\n\x0fGET_SUPER_TYPES\x10\x13\x12\x18\n\x14GET_TABLE_PRIVILEGES\x10\x14\x12\x0e\n\nGET_TABLES\x10\x15\x12\x13\n\x0fGET_TABLE_TYPES\x10\x16\x12\x11\n\rGET_TYPE_INFO\x10\x17\x12\x0c\n\x08GET_UDTS\x10\x18\x12\x17\n\x13GET_VERSION_COLUMNS\x10\x19*\"\n\tStateType\x12\x07\n\x03SQL\x10\x00\x12\x0c\n\x08METADATA\x10\x01\x42\"\n org.apache.calcite.avatica.protob\x06proto3')

_STATEMENTTYPE = DESCRIPTOR.enum_types_by_name['StatementType']
StatementType = enum_type_wrapper.EnumTypeWrapper(_STATEMENTTYPE)
_REP = DESCRIPTOR.enum_types_by_name['Rep']
Rep = enum_type_wrapper.EnumTypeWrapper(_REP)
_SEVERITY = DESCRIPTOR.enum_types_by_name['Severity']
Severity = enum_type_wrapper.EnumTypeWrapper(_SEVERITY)
_METADATAOPERATION = DESCRIPTOR.enum_types_by_name['MetaDataOperation']
MetaDataOperation = enum_type_wrapper.EnumTypeWrapper(_METADATAOPERATION)
_STATETYPE = DESCRIPTOR.enum_types_by_name['StateType']
StateType = enum_type_wrapper.EnumTypeWrapper(_STATETYPE)
SELECT = 0
INSERT = 1
UPDATE = 2
DELETE = 3
UPSERT = 4
MERGE = 5
OTHER_DML = 6
CREATE = 7
DROP = 8
ALTER = 9
OTHER_DDL = 10
CALL = 11
PRIMITIVE_BOOLEAN = 0
PRIMITIVE_BYTE = 1
PRIMITIVE_CHAR = 2
PRIMITIVE_SHORT = 3
PRIMITIVE_INT = 4
PRIMITIVE_LONG = 5
PRIMITIVE_FLOAT = 6
PRIMITIVE_DOUBLE = 7
BOOLEAN = 8
BYTE = 9
CHARACTER = 10
SHORT = 11
INTEGER = 12
LONG = 13
FLOAT = 14
DOUBLE = 15
BIG_INTEGER = 25
BIG_DECIMAL = 26
JAVA_SQL_TIME = 16
JAVA_SQL_TIMESTAMP = 17
JAVA_SQL_DATE = 18
JAVA_UTIL_DATE = 19
BYTE_STRING = 20
STRING = 21
NUMBER = 22
OBJECT = 23
NULL = 24
ARRAY = 27
STRUCT = 28
MULTISET = 29
UNKNOWN_SEVERITY = 0
FATAL_SEVERITY = 1
ERROR_SEVERITY = 2
WARNING_SEVERITY = 3
GET_ATTRIBUTES = 0
GET_BEST_ROW_IDENTIFIER = 1
GET_CATALOGS = 2
GET_CLIENT_INFO_PROPERTIES = 3
GET_COLUMN_PRIVILEGES = 4
GET_COLUMNS = 5
GET_CROSS_REFERENCE = 6
GET_EXPORTED_KEYS = 7
GET_FUNCTION_COLUMNS = 8
GET_FUNCTIONS = 9
GET_IMPORTED_KEYS = 10
GET_INDEX_INFO = 11
GET_PRIMARY_KEYS = 12
GET_PROCEDURE_COLUMNS = 13
GET_PROCEDURES = 14
GET_PSEUDO_COLUMNS = 15
GET_SCHEMAS = 16
GET_SCHEMAS_WITH_ARGS = 17
GET_SUPER_TABLES = 18
GET_SUPER_TYPES = 19
GET_TABLE_PRIVILEGES = 20
GET_TABLES = 21
GET_TABLE_TYPES = 22
GET_TYPE_INFO = 23
GET_UDTS = 24
GET_VERSION_COLUMNS = 25
SQL = 0
METADATA = 1


_CONNECTIONPROPERTIES = DESCRIPTOR.message_types_by_name['ConnectionProperties']
_STATEMENTHANDLE = DESCRIPTOR.message_types_by_name['StatementHandle']
_SIGNATURE = DESCRIPTOR.message_types_by_name['Signature']
_COLUMNMETADATA = DESCRIPTOR.message_types_by_name['ColumnMetaData']
_AVATICATYPE = DESCRIPTOR.message_types_by_name['AvaticaType']
_AVATICAPARAMETER = DESCRIPTOR.message_types_by_name['AvaticaParameter']
_CURSORFACTORY = DESCRIPTOR.message_types_by_name['CursorFactory']
_FRAME = DESCRIPTOR.message_types_by_name['Frame']
_ROW = DESCRIPTOR.message_types_by_name['Row']
_DATABASEPROPERTY = DESCRIPTOR.message_types_by_name['DatabaseProperty']
_WIREMESSAGE = DESCRIPTOR.message_types_by_name['WireMessage']
_COLUMNVALUE = DESCRIPTOR.message_types_by_name['ColumnValue']
_TYPEDVALUE = DESCRIPTOR.message_types_by_name['TypedValue']
_METADATAOPERATIONARGUMENT = DESCRIPTOR.message_types_by_name['MetaDataOperationArgument']
_QUERYSTATE = DESCRIPTOR.message_types_by_name['QueryState']
_KEYVALUEPAIR = DESCRIPTOR.message_types_by_name['KeyValuePair']
_CURSORFACTORY_STYLE = _CURSORFACTORY.enum_types_by_name['Style']
_METADATAOPERATIONARGUMENT_ARGUMENTTYPE = _METADATAOPERATIONARGUMENT.enum_types_by_name['ArgumentType']
ConnectionProperties = _reflection.GeneratedProtocolMessageType('ConnectionProperties', (_message.Message,), {
  'DESCRIPTOR' : _CONNECTIONPROPERTIES,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.ConnectionProperties)
  })
_sym_db.RegisterMessage(ConnectionProperties)

StatementHandle = _reflection.GeneratedProtocolMessageType('StatementHandle', (_message.Message,), {
  'DESCRIPTOR' : _STATEMENTHANDLE,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.StatementHandle)
  })
_sym_db.RegisterMessage(StatementHandle)

Signature = _reflection.GeneratedProtocolMessageType('Signature', (_message.Message,), {
  'DESCRIPTOR' : _SIGNATURE,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.Signature)
  })
_sym_db.RegisterMessage(Signature)

ColumnMetaData = _reflection.GeneratedProtocolMessageType('ColumnMetaData', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNMETADATA,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.ColumnMetaData)
  })
_sym_db.RegisterMessage(ColumnMetaData)

AvaticaType = _reflection.GeneratedProtocolMessageType('AvaticaType', (_message.Message,), {
  'DESCRIPTOR' : _AVATICATYPE,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.AvaticaType)
  })
_sym_db.RegisterMessage(AvaticaType)

AvaticaParameter = _reflection.GeneratedProtocolMessageType('AvaticaParameter', (_message.Message,), {
  'DESCRIPTOR' : _AVATICAPARAMETER,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.AvaticaParameter)
  })
_sym_db.RegisterMessage(AvaticaParameter)

CursorFactory = _reflection.GeneratedProtocolMessageType('CursorFactory', (_message.Message,), {
  'DESCRIPTOR' : _CURSORFACTORY,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.CursorFactory)
  })
_sym_db.RegisterMessage(CursorFactory)

Frame = _reflection.GeneratedProtocolMessageType('Frame', (_message.Message,), {
  'DESCRIPTOR' : _FRAME,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.Frame)
  })
_sym_db.RegisterMessage(Frame)

Row = _reflection.GeneratedProtocolMessageType('Row', (_message.Message,), {
  'DESCRIPTOR' : _ROW,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.Row)
  })
_sym_db.RegisterMessage(Row)

DatabaseProperty = _reflection.GeneratedProtocolMessageType('DatabaseProperty', (_message.Message,), {
  'DESCRIPTOR' : _DATABASEPROPERTY,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.DatabaseProperty)
  })
_sym_db.RegisterMessage(DatabaseProperty)

WireMessage = _reflection.GeneratedProtocolMessageType('WireMessage', (_message.Message,), {
  'DESCRIPTOR' : _WIREMESSAGE,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.WireMessage)
  })
_sym_db.RegisterMessage(WireMessage)

ColumnValue = _reflection.GeneratedProtocolMessageType('ColumnValue', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNVALUE,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.ColumnValue)
  })
_sym_db.RegisterMessage(ColumnValue)

TypedValue = _reflection.GeneratedProtocolMessageType('TypedValue', (_message.Message,), {
  'DESCRIPTOR' : _TYPEDVALUE,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.TypedValue)
  })
_sym_db.RegisterMessage(TypedValue)

MetaDataOperationArgument = _reflection.GeneratedProtocolMessageType('MetaDataOperationArgument', (_message.Message,), {
  'DESCRIPTOR' : _METADATAOPERATIONARGUMENT,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.MetaDataOperationArgument)
  })
_sym_db.RegisterMessage(MetaDataOperationArgument)

QueryState = _reflection.GeneratedProtocolMessageType('QueryState', (_message.Message,), {
  'DESCRIPTOR' : _QUERYSTATE,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.QueryState)
  })
_sym_db.RegisterMessage(QueryState)

KeyValuePair = _reflection.GeneratedProtocolMessageType('KeyValuePair', (_message.Message,), {
  'DESCRIPTOR' : _KEYVALUEPAIR,
  '__module__' : 'lxdbapi.common_pb2'
  # @@protoc_insertion_point(class_scope:lxdbapi.KeyValuePair)
  })
_sym_db.RegisterMessage(KeyValuePair)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n org.apache.calcite.avatica.proto'
  _STATEMENTTYPE._serialized_start=2685
  _STATEMENTTYPE._serialized_end=2844
  _REP._serialized_start=2847
  _REP._serialized_end=3329
  _SEVERITY._serialized_start=3331
  _SEVERITY._serialized_end=3425
  _METADATAOPERATION._serialized_start=3428
  _METADATAOPERATION._serialized_end=4027
  _STATETYPE._serialized_start=4029
  _STATETYPE._serialized_end=4063
  _CONNECTIONPROPERTIES._serialized_start=34
  _CONNECTIONPROPERTIES._serialized_end=226
  _STATEMENTHANDLE._serialized_start=228
  _STATEMENTHANDLE._serialized_end=319
  _SIGNATURE._serialized_start=322
  _SIGNATURE._serialized_end=530
  _COLUMNMETADATA._serialized_start=533
  _COLUMNMETADATA._serialized_end=970
  _AVATICATYPE._serialized_start=973
  _AVATICATYPE._serialized_end=1122
  _AVATICAPARAMETER._serialized_start=1125
  _AVATICAPARAMETER._serialized_end=1270
  _CURSORFACTORY._serialized_start=1273
  _CURSORFACTORY._serialized_end=1460
  _CURSORFACTORY_STYLE._serialized_start=1376
  _CURSORFACTORY_STYLE._serialized_end=1460
  _FRAME._serialized_start=1462
  _FRAME._serialized_end=1527
  _ROW._serialized_start=1529
  _ROW._serialized_end=1571
  _DATABASEPROPERTY._serialized_start=1573
  _DATABASEPROPERTY._serialized_end=1624
  _WIREMESSAGE._serialized_start=1626
  _WIREMESSAGE._serialized_end=1678
  _COLUMNVALUE._serialized_start=1681
  _COLUMNVALUE._serialized_end=1840
  _TYPEDVALUE._serialized_start=1843
  _TYPEDVALUE._serialized_end=2109
  _METADATAOPERATIONARGUMENT._serialized_start=2112
  _METADATAOPERATIONARGUMENT._serialized_end=2414
  _METADATAOPERATIONARGUMENT_ARGUMENTTYPE._serialized_start=2320
  _METADATAOPERATIONARGUMENT_ARGUMENTTYPE._serialized_end=2414
  _QUERYSTATE._serialized_start=2417
  _QUERYSTATE._serialized_end=2617
  _KEYVALUEPAIR._serialized_start=2619
  _KEYVALUEPAIR._serialized_end=2682
# @@protoc_insertion_point(module_scope)



"""
For the ``future`` package.

Adds this import line::

    from builtins import XYZ

for each of the functions XYZ that is used in the module.

Adds these imports after any other imports (in an initial block of them).
"""

from __future__ import unicode_literals

from lib2to3 import fixer_base
from lib2to3.pygram import python_symbols as syms
from lib2to3.fixer_util import Name, Call, in_special_context

from ...libfuturize.fixer_util import touch_import_top

# All builtins are:
#     from future.builtins.iterators import (filter, map, zip)
#     from future.builtins.misc import (ascii, chr, hex, input, isinstance, oct, open, round, super)
#     from future.types import (bytes, dict, int, range, str)
# We don't need isinstance any more.

replaced_builtin_fns = '''filter map zip
                       ascii chr hex input next oct
                       bytes range str raw_input'''.split()
                       # This includes raw_input as a workaround for the
                       # lib2to3 fixer for raw_input on Py3 (only), allowing
                       # the correct import to be included. (Py3 seems to run
                       # the fixers the wrong way around, perhaps ignoring the
                       # run_order class attribute below ...)

expression = '|'.join(["name='{0}'".format(name) for name in replaced_builtin_fns])


class FixFutureBuiltins(fixer_base.BaseFix):
    BM_compatible = True
    run_order = 7

    # Currently we only match uses as a function. This doesn't match e.g.:
    #     if isinstance(s, str):
    #         ...
    PATTERN = """
              power<
                 ({0}) trailer< '(' [arglist=any] ')' >
              rest=any* >
              |
              power<
                  'map' trailer< '(' [arglist=any] ')' >
              >
              """.format(expression)

    def transform(self, node, results):
        name = results["name"]
        touch_import_top(u'builtins', name.value, node)
        # name.replace(Name(u"input", prefix=name.prefix))



# Copyright 2015 Lukas Lalinsky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from lxdbapi import errors
from lxdbapi import typessOld
from lxdbapi.multi_connection import MultiConnection
from lxdbapi.errors import *
from lxdbapi.typessOld import *

__all__ = ['connect', 'apilevel', 'threadsafety', 'paramstyle'] + typessOld.__all__ + errors.__all__


apilevel = "2.0"
"""
This module supports the `DB API 2.0 interface <https://www.python.org/dev/peps/pep-0249/>`_.
"""

threadsafety = 1
"""
Multiple threads can share the module, but neither connections nor cursors.
"""

paramstyle = 'qmark'
"""
Parmetrized queries should use the question mark as a parameter placeholder.

For example::

 cursor.execute("SELECT * FROM table WHERE id = ?", [my_id])
"""


def connect(url, max_retries=None, **kwargs):
    """Connects to a LeanXcale query server.

    :param url:
        URL to the Phoenix query server, e.g. ``http://localhost:8765/``

    :param autocommit:
        Switch the connection to autocommit mode.

    :param readonly:
        Switch the connection to readonly mode.

    :param max_retries:
        The maximum number of retries in case there is a connection error.

    :param cursor_factory:
        If specified, the connection's :attr:`~phoenixdb.connection.Connection.cursor_factory` is set to it.

    :returns:
        :class:`~phoenixdb.connection.Connection` object.
    """
    info = kwargs
    url = url.replace('localhost', '127.0.0.1')
    result = MultiConnection(url, info)
    return result



# coding=utf-8
"""
past: compatibility with Python 2 from Python 3
===============================================

``past`` is a package to aid with Python 2/3 compatibility. Whereas ``future``
contains backports of Python 3 constructs to Python 2, ``past`` provides
implementations of some Python 2 constructs in Python 3 and tools to import and
run Python 2 code in Python 3. It is intended to be used sparingly, as a way of
running old Python 2 code from Python 3 until the code is ported properly.

Potential uses for libraries:

- as a step in porting a Python 2 codebase to Python 3 (e.g. with the ``futurize`` script)
- to provide Python 3 support for previously Python 2-only libraries with the
  same APIs as on Python 2 -- particularly with regard to 8-bit strings (the
  ``past.builtins.str`` type).
- to aid in providing minimal-effort Python 3 support for applications using
  libraries that do not yet wish to upgrade their code properly to Python 3, or
  wish to upgrade it gradually to Python 3 style.


Here are some code examples that run identically on Python 3 and 2::

    >>> from past.builtins import str as oldstr

    >>> philosopher = oldstr(u'\u5b54\u5b50'.encode('utf-8'))
    >>> # This now behaves like a Py2 byte-string on both Py2 and Py3.
    >>> # For example, indexing returns a Python 2-like string object, not
    >>> # an integer:
    >>> philosopher[0]
    '\xe5'
    >>> type(philosopher[0])
    <past.builtins.oldstr>

    >>> # List-producing versions of range, reduce, map, filter
    >>> from past.builtins import range, reduce
    >>> range(10)
    [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    >>> reduce(lambda x, y: x+y, [1, 2, 3, 4, 5])
    15

    >>> # Other functions removed in Python 3 are resurrected ...
    >>> from past.builtins import execfile
    >>> execfile('myfile.py')

    >>> from past.builtins import raw_input
    >>> name = raw_input('What is your name? ')
    What is your name? [cursor]

    >>> from past.builtins import reload
    >>> reload(mymodule)   # equivalent to imp.reload(mymodule) in Python 3

    >>> from past.builtins import xrange
    >>> for i in xrange(10):
    ...     pass


It also provides import hooks so you can import and use Python 2 modules like
this::

    $ python3

    >>> from past.translation import autotranslate
    >>> authotranslate('mypy2module')
    >>> import mypy2module

until the authors of the Python 2 modules have upgraded their code. Then, for
example::

    >>> mypy2module.func_taking_py2_string(oldstr(b'abcd'))


Credits
-------

:Author:  Ed Schofield, Jordan M. Adler, et al
:Sponsor: Python Charmers Pty Ltd, Australia: http://pythoncharmers.com


Licensing
---------
Copyright 2013-2019 Python Charmers Pty Ltd, Australia.
The software is distributed under an MIT licence. See LICENSE.txt.
"""

from ..future import __version__, __copyright__, __license__

__title__ = 'past'
__author__ = 'Ed Schofield'



import logging

import math
import socket
from threading import RLock

from lxdbapi import errors, versions
import ssl
import time

try:
    import httplib
except ImportError:
    import http.client as httplib

__all__ = ['HttpAvaticaConnection', 'TcpAvaticaConnection', 'NetAvaticaConnection']

logger = logging.getLogger(__name__)

class NetAvaticaConnection(object):
    def close(self):
        raise NotImplementedError('Extend NetAvaticaConnection')

    def request(self, body=None):
        raise NotImplementedError('Extend NetAvaticaConnection')

    def checkSerialization(self):
        raise NotImplementedError('Extend NetAvaticaConnection')


class TcpAvaticaConnection(NetAvaticaConnection):

    def __init__(self, url, secure, max_retries):
        self._close_lock = RLock()
        self.url = url
        self.secure =secure
        """Opens a HTTP connection to the RPC server."""
        self.max_retries = max_retries if max_retries is not None else 3
        self._opened = False
        self._connect()

    def __del__(self):
        self._close()

    def __exit__(self):
        self._close()

    def _connect(self):
        if not self.secure:
            logger.debug("Using TCP")
            logger.debug("Opening connection to %s:%s", self.url.hostname, self.url.port)
            self.tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.tcp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.tcp_socket.connect((self.url.hostname, self.url.port))
            self.tcpConn = self.tcp_socket
            self._opened = True
        else:
            logger.debug("Using TCP with SSL")
            logger.debug("Opening connection to %s:%s", self.url.hostname, self.url.port)
            self.tcp_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.tcp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
            self.tcpConn = ssl.wrap_socket(self.tcp_socket, ssl_version=ssl.PROTOCOL_TLSv1)
            self.tcpConn.connect((self.url.hostname, self.url.port))
            self._opened = True

    def _close(self):
        with self._close_lock:
            if self._opened:
                self._opened = False
                self.tcpConn.close()

    def close(self):
        self._close()

    def request(self, body=None):
        req_sz = len(body)
        # build message Body size, tag(always 0), body
        tag = 0
        msg = bytearray()
        msg.extend(req_sz.to_bytes(4, 'little'))
        msg.extend(tag.to_bytes(4, 'little'))
        msg.extend(body)
        self.tcpConn.sendall(msg)
        hdr = self.tcpConn.recv(8)
        resp_sz = int.from_bytes(hdr[:4], 'little')
        tag = int.from_bytes(hdr[4:], 'little')
        if 0 > resp_sz:
            logger.warning("negative msg size (sz {}) tag:{}".format(resp_sz, tag))
            raise IOError('IO Error on RPC request on {}. Got negative size [{}]'.format(self.url, resp_sz))
        if 0 == resp_sz:
            return None
        response = bytearray()
        pending = resp_sz
        offset = 0
        while pending > 0:
            partial = self.tcpConn.recv(resp_sz)
            l = len(partial)
            offset += l
            pending -= l
            response.extend(partial)
        return response

    def checkSerialization(self):
        raise NotImplementedError('Extend NetAvaticaConnection')


class HttpAvaticaConnection(NetAvaticaConnection):

    def __init__(self, url, secure, max_retries):
        self.url = url
        self.secure =secure
        """Opens a HTTP connection to the RPC server."""
        self.max_retries = max_retries if max_retries is not None else 3
        self._connect()

    def _connect(self):
        if not self.secure:
            logger.debug("Using HTTP")
            logger.debug("Opening connection to %s:%s", self.url.hostname, self.url.port)
            try:
                self.httpConn = httplib.HTTPConnection(self.url.hostname, self.url.port)
                self.httpConn.connect()
                self.cookies = {}
                # self.checkSerialization()
            except (httplib.HTTPException, socket.error) as e:
                raise errors.InterfaceError('Unable to connect to the specified service', e)
        else:
            logger.debug("Using HTTPS")
            logger.debug("Opening connection to %s:%s", self.url.hostname, self.url.port)
            try:
                purpose = ssl.Purpose.SERVER_AUTH
                context = ssl.create_default_context(purpose)
                self.httpConn = httplib.HTTPSConnection(self.url.hostname, self.url.port,
                                                        context=context)
                self.httpConn.connect()
                self.cookies = {}
                # self.checkSerialization()
            except (httplib.HTTPException, socket.error) as e:
                raise errors.InterfaceError('Unable to connect to the specified service', e)

    def close(self):
        """Closes the HTTP connection to the RPC server."""
        logger.debug("Closing connection to %s:%s", self.url.hostname, self.url.port)
        try:
            self.httpConn.close()
        except httplib.HTTPException:
            logger.warning("Error while closing connection", exc_info=True)

    def request(self, body=None):
        method = 'POST'
        headers = {'content-type': 'application/x-google-protobuf'}
        # add cookies if any
        if self.cookies is not None and self.cookies.get('hap-cookie') is not None:
            headers['Cookie'] = self.cookies.get('hap-cookie')
        retry_count = self.max_retries
        while True:
            logger.debug("%s %s %r %r", method, self.url.path, body, headers)
            try:
                self.httpConn.request(method, self.url.path, body=body, headers=headers)
                response = self.httpConn.getresponse()
            except (httplib.HTTPException, ssl.SSLError, BrokenPipeError) as e:
                if retry_count > 0:
                    logger.debug("Error at HTTP connection level... Trying to reconnect to {}".format(self.url))
                    delay = math.exp(-retry_count)
                    logger.debug("HTTP protocol error, will retry in %s seconds...", delay, exc_info=True)
                    self.close()
                    try:
                        self._connect()
                    except BaseException as th:
                        logger.debug("Retry failed on {}: {}".format(self.url, th))
                        raise th
                    logger.debug("Successfully reconnected HTTP connection, Continuing...")
                    time.sleep(delay)
                    retry_count -= 1
                    continue
                raise errors.InterfaceError('RPC request failed on {}'.format(self.url), cause=e)
            else:
                if response.status == httplib.SERVICE_UNAVAILABLE:
                    if retry_count > 0:
                        delay = math.exp(-retry_count)
                        logger.debug("Service unavailable, will retry in %s seconds...", delay, exc_info=True)
                        time.sleep(delay)
                        retry_count -= 1
                        continue
                # save cookie if using hap
                if response.headers.get('set-cookie') is not None:
                    self.cookies['hap-cookie'] = response.headers.get('set-cookie')
                    logger.debug("Setting cookie: {}:{}".format('hap-cookie', response.headers.get('set-cookie')))
                return response.read()

    def checkSerialization(self):
        retry_count = self.max_retries
        while True:
            logger.debug("OPTIONS %s", self.url.path)
            try:
                self.httpConn.request('OPTIONS', '')
                response = self.httpConn.getresponse()
                response_body = response.read()
                serialization = str(response_body).split(';')

                # Compare the minimum server version against client version
                index = [idx for idx, s in enumerate(serialization) if 'minimumVersion' in s]
                if index:
                    minimumVersions = serialization[index[0]].split("=")[1].split(".")
                    if int(minimumVersions[0]) > int(versions.MAJOR_VERSION) or \
                            (int(minimumVersions[0]) == int(versions.MAJOR_VERSION) and int(minimumVersions[1]) > int(
                                versions.MINOR_VERSION)):
                        raise errors.SerializationError(
                            "Client version '{}.{}' is not allowed to connect. Minimum allowed version is '{}.{}'".format(
                                versions.MAJOR_VERSION, versions.MINOR_VERSION, minimumVersions[0], minimumVersions[1]))

            except IndexError as e:
                raise errors.InterfaceError('Avatica server did not respond correctly. Please check endpoint.', cause=e)
            except httplib.HTTPException as e:
                if retry_count > 0:
                    retry_count -= 1
                    delay = math.exp(-retry_count)
                    logger.debug("HTTP protocol error, will retry in %s seconds...", delay, exc_info=True)
                    self.close()
                    self._connect()
                    time.sleep(delay)
                    continue
                raise errors.InterfaceError('Max retries for attempting to connect and check serialization wasted',
                                            cause=e)
            else:
                if response.status == httplib.SERVICE_UNAVAILABLE:
                    if retry_count > 0:
                        delay = math.exp(-retry_count)
                        logger.debug("Service unavailable, will retry in %s seconds...", delay, exc_info=True)
                        time.sleep(delay)
                        retry_count -= 1
                        continue
                return response



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: google/protobuf/util/json_format_proto3.proto
"""Generated protocol buffer code."""
from ....google.protobuf.internal import enum_type_wrapper
from ....google.protobuf import descriptor as _descriptor
from ....google.protobuf import descriptor_pool as _descriptor_pool
from ....google.protobuf import message as _message
from ....google.protobuf import reflection as _reflection
from ....google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ....google.protobuf import any_pb2 as google_dot_protobuf_dot_any__pb2
from ....google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2
from ....google.protobuf import field_mask_pb2 as google_dot_protobuf_dot_field__mask__pb2
from ....google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from ....google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from ....google.protobuf import wrappers_pb2 as google_dot_protobuf_dot_wrappers__pb2
from ....google.protobuf import unittest_pb2 as google_dot_protobuf_dot_unittest__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n-google/protobuf/util/json_format_proto3.proto\x12\x06proto3\x1a\x19google/protobuf/any.proto\x1a\x1egoogle/protobuf/duration.proto\x1a google/protobuf/field_mask.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1egoogle/protobuf/wrappers.proto\x1a\x1egoogle/protobuf/unittest.proto\"\x1c\n\x0bMessageType\x12\r\n\x05value\x18\x01 \x01(\x05\"\x94\x05\n\x0bTestMessage\x12\x12\n\nbool_value\x18\x01 \x01(\x08\x12\x13\n\x0bint32_value\x18\x02 \x01(\x05\x12\x13\n\x0bint64_value\x18\x03 \x01(\x03\x12\x14\n\x0cuint32_value\x18\x04 \x01(\r\x12\x14\n\x0cuint64_value\x18\x05 \x01(\x04\x12\x13\n\x0b\x66loat_value\x18\x06 \x01(\x02\x12\x14\n\x0c\x64ouble_value\x18\x07 \x01(\x01\x12\x14\n\x0cstring_value\x18\x08 \x01(\t\x12\x13\n\x0b\x62ytes_value\x18\t \x01(\x0c\x12$\n\nenum_value\x18\n \x01(\x0e\x32\x10.proto3.EnumType\x12*\n\rmessage_value\x18\x0b \x01(\x0b\x32\x13.proto3.MessageType\x12\x1b\n\x13repeated_bool_value\x18\x15 \x03(\x08\x12\x1c\n\x14repeated_int32_value\x18\x16 \x03(\x05\x12\x1c\n\x14repeated_int64_value\x18\x17 \x03(\x03\x12\x1d\n\x15repeated_uint32_value\x18\x18 \x03(\r\x12\x1d\n\x15repeated_uint64_value\x18\x19 \x03(\x04\x12\x1c\n\x14repeated_float_value\x18\x1a \x03(\x02\x12\x1d\n\x15repeated_double_value\x18\x1b \x03(\x01\x12\x1d\n\x15repeated_string_value\x18\x1c \x03(\t\x12\x1c\n\x14repeated_bytes_value\x18\x1d \x03(\x0c\x12-\n\x13repeated_enum_value\x18\x1e \x03(\x0e\x32\x10.proto3.EnumType\x12\x33\n\x16repeated_message_value\x18\x1f \x03(\x0b\x32\x13.proto3.MessageType\"\x8c\x02\n\tTestOneof\x12\x1b\n\x11oneof_int32_value\x18\x01 \x01(\x05H\x00\x12\x1c\n\x12oneof_string_value\x18\x02 \x01(\tH\x00\x12\x1b\n\x11oneof_bytes_value\x18\x03 \x01(\x0cH\x00\x12,\n\x10oneof_enum_value\x18\x04 \x01(\x0e\x32\x10.proto3.EnumTypeH\x00\x12\x32\n\x13oneof_message_value\x18\x05 \x01(\x0b\x32\x13.proto3.MessageTypeH\x00\x12\x36\n\x10oneof_null_value\x18\x06 \x01(\x0e\x32\x1a.google.protobuf.NullValueH\x00\x42\r\n\x0boneof_value\"\xe1\x04\n\x07TestMap\x12.\n\x08\x62ool_map\x18\x01 \x03(\x0b\x32\x1c.proto3.TestMap.BoolMapEntry\x12\x30\n\tint32_map\x18\x02 \x03(\x0b\x32\x1d.proto3.TestMap.Int32MapEntry\x12\x30\n\tint64_map\x18\x03 \x03(\x0b\x32\x1d.proto3.TestMap.Int64MapEntry\x12\x32\n\nuint32_map\x18\x04 \x03(\x0b\x32\x1e.proto3.TestMap.Uint32MapEntry\x12\x32\n\nuint64_map\x18\x05 \x03(\x0b\x32\x1e.proto3.TestMap.Uint64MapEntry\x12\x32\n\nstring_map\x18\x06 \x03(\x0b\x32\x1e.proto3.TestMap.StringMapEntry\x1a.\n\x0c\x42oolMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\x08\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a/\n\rInt32MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a/\n\rInt64MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x30\n\x0eUint32MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x30\n\x0eUint64MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\x04\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x30\n\x0eStringMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\"\x85\x06\n\rTestNestedMap\x12\x34\n\x08\x62ool_map\x18\x01 \x03(\x0b\x32\".proto3.TestNestedMap.BoolMapEntry\x12\x36\n\tint32_map\x18\x02 \x03(\x0b\x32#.proto3.TestNestedMap.Int32MapEntry\x12\x36\n\tint64_map\x18\x03 \x03(\x0b\x32#.proto3.TestNestedMap.Int64MapEntry\x12\x38\n\nuint32_map\x18\x04 \x03(\x0b\x32$.proto3.TestNestedMap.Uint32MapEntry\x12\x38\n\nuint64_map\x18\x05 \x03(\x0b\x32$.proto3.TestNestedMap.Uint64MapEntry\x12\x38\n\nstring_map\x18\x06 \x03(\x0b\x32$.proto3.TestNestedMap.StringMapEntry\x12\x32\n\x07map_map\x18\x07 \x03(\x0b\x32!.proto3.TestNestedMap.MapMapEntry\x1a.\n\x0c\x42oolMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\x08\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a/\n\rInt32MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a/\n\rInt64MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x30\n\x0eUint32MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x30\n\x0eUint64MapEntry\x12\x0b\n\x03key\x18\x01 \x01(\x04\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x30\n\x0eStringMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\x1a\x44\n\x0bMapMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12$\n\x05value\x18\x02 \x01(\x0b\x32\x15.proto3.TestNestedMap:\x02\x38\x01\"{\n\rTestStringMap\x12\x38\n\nstring_map\x18\x01 \x03(\x0b\x32$.proto3.TestStringMap.StringMapEntry\x1a\x30\n\x0eStringMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\xee\x07\n\x0bTestWrapper\x12.\n\nbool_value\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x30\n\x0bint32_value\x18\x02 \x01(\x0b\x32\x1b.google.protobuf.Int32Value\x12\x30\n\x0bint64_value\x18\x03 \x01(\x0b\x32\x1b.google.protobuf.Int64Value\x12\x32\n\x0cuint32_value\x18\x04 \x01(\x0b\x32\x1c.google.protobuf.UInt32Value\x12\x32\n\x0cuint64_value\x18\x05 \x01(\x0b\x32\x1c.google.protobuf.UInt64Value\x12\x30\n\x0b\x66loat_value\x18\x06 \x01(\x0b\x32\x1b.google.protobuf.FloatValue\x12\x32\n\x0c\x64ouble_value\x18\x07 \x01(\x0b\x32\x1c.google.protobuf.DoubleValue\x12\x32\n\x0cstring_value\x18\x08 \x01(\x0b\x32\x1c.google.protobuf.StringValue\x12\x30\n\x0b\x62ytes_value\x18\t \x01(\x0b\x32\x1b.google.protobuf.BytesValue\x12\x37\n\x13repeated_bool_value\x18\x0b \x03(\x0b\x32\x1a.google.protobuf.BoolValue\x12\x39\n\x14repeated_int32_value\x18\x0c \x03(\x0b\x32\x1b.google.protobuf.Int32Value\x12\x39\n\x14repeated_int64_value\x18\r \x03(\x0b\x32\x1b.google.protobuf.Int64Value\x12;\n\x15repeated_uint32_value\x18\x0e \x03(\x0b\x32\x1c.google.protobuf.UInt32Value\x12;\n\x15repeated_uint64_value\x18\x0f \x03(\x0b\x32\x1c.google.protobuf.UInt64Value\x12\x39\n\x14repeated_float_value\x18\x10 \x03(\x0b\x32\x1b.google.protobuf.FloatValue\x12;\n\x15repeated_double_value\x18\x11 \x03(\x0b\x32\x1c.google.protobuf.DoubleValue\x12;\n\x15repeated_string_value\x18\x12 \x03(\x0b\x32\x1c.google.protobuf.StringValue\x12\x39\n\x14repeated_bytes_value\x18\x13 \x03(\x0b\x32\x1b.google.protobuf.BytesValue\"n\n\rTestTimestamp\x12)\n\x05value\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x32\n\x0erepeated_value\x18\x02 \x03(\x0b\x32\x1a.google.protobuf.Timestamp\"k\n\x0cTestDuration\x12(\n\x05value\x18\x01 \x01(\x0b\x32\x19.google.protobuf.Duration\x12\x31\n\x0erepeated_value\x18\x02 \x03(\x0b\x32\x19.google.protobuf.Duration\":\n\rTestFieldMask\x12)\n\x05value\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.FieldMask\"e\n\nTestStruct\x12&\n\x05value\x18\x01 \x01(\x0b\x32\x17.google.protobuf.Struct\x12/\n\x0erepeated_value\x18\x02 \x03(\x0b\x32\x17.google.protobuf.Struct\"\\\n\x07TestAny\x12#\n\x05value\x18\x01 \x01(\x0b\x32\x14.google.protobuf.Any\x12,\n\x0erepeated_value\x18\x02 \x03(\x0b\x32\x14.google.protobuf.Any\"b\n\tTestValue\x12%\n\x05value\x18\x01 \x01(\x0b\x32\x16.google.protobuf.Value\x12.\n\x0erepeated_value\x18\x02 \x03(\x0b\x32\x16.google.protobuf.Value\"n\n\rTestListValue\x12)\n\x05value\x18\x01 \x01(\x0b\x32\x1a.google.protobuf.ListValue\x12\x32\n\x0erepeated_value\x18\x02 \x03(\x0b\x32\x1a.google.protobuf.ListValue\"\x89\x01\n\rTestBoolValue\x12\x12\n\nbool_value\x18\x01 \x01(\x08\x12\x34\n\x08\x62ool_map\x18\x02 \x03(\x0b\x32\".proto3.TestBoolValue.BoolMapEntry\x1a.\n\x0c\x42oolMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\x08\x12\r\n\x05value\x18\x02 \x01(\x05:\x02\x38\x01\"+\n\x12TestCustomJsonName\x12\x15\n\x05value\x18\x01 \x01(\x05R\x06@value\"J\n\x0eTestExtensions\x12\x38\n\nextensions\x18\x01 \x01(\x0b\x32$.protobuf_unittest.TestAllExtensions\"\x84\x01\n\rTestEnumValue\x12%\n\x0b\x65num_value1\x18\x01 \x01(\x0e\x32\x10.proto3.EnumType\x12%\n\x0b\x65num_value2\x18\x02 \x01(\x0e\x32\x10.proto3.EnumType\x12%\n\x0b\x65num_value3\x18\x03 \x01(\x0e\x32\x10.proto3.EnumType*\x1c\n\x08\x45numType\x12\x07\n\x03\x46OO\x10\x00\x12\x07\n\x03\x42\x41R\x10\x01\x42,\n\x18\x63om.google.protobuf.utilB\x10JsonFormatProto3b\x06proto3')

_ENUMTYPE = DESCRIPTOR.enum_types_by_name['EnumType']
EnumType = enum_type_wrapper.EnumTypeWrapper(_ENUMTYPE)
FOO = 0
BAR = 1


_MESSAGETYPE = DESCRIPTOR.message_types_by_name['MessageType']
_TESTMESSAGE = DESCRIPTOR.message_types_by_name['TestMessage']
_TESTONEOF = DESCRIPTOR.message_types_by_name['TestOneof']
_TESTMAP = DESCRIPTOR.message_types_by_name['TestMap']
_TESTMAP_BOOLMAPENTRY = _TESTMAP.nested_types_by_name['BoolMapEntry']
_TESTMAP_INT32MAPENTRY = _TESTMAP.nested_types_by_name['Int32MapEntry']
_TESTMAP_INT64MAPENTRY = _TESTMAP.nested_types_by_name['Int64MapEntry']
_TESTMAP_UINT32MAPENTRY = _TESTMAP.nested_types_by_name['Uint32MapEntry']
_TESTMAP_UINT64MAPENTRY = _TESTMAP.nested_types_by_name['Uint64MapEntry']
_TESTMAP_STRINGMAPENTRY = _TESTMAP.nested_types_by_name['StringMapEntry']
_TESTNESTEDMAP = DESCRIPTOR.message_types_by_name['TestNestedMap']
_TESTNESTEDMAP_BOOLMAPENTRY = _TESTNESTEDMAP.nested_types_by_name['BoolMapEntry']
_TESTNESTEDMAP_INT32MAPENTRY = _TESTNESTEDMAP.nested_types_by_name['Int32MapEntry']
_TESTNESTEDMAP_INT64MAPENTRY = _TESTNESTEDMAP.nested_types_by_name['Int64MapEntry']
_TESTNESTEDMAP_UINT32MAPENTRY = _TESTNESTEDMAP.nested_types_by_name['Uint32MapEntry']
_TESTNESTEDMAP_UINT64MAPENTRY = _TESTNESTEDMAP.nested_types_by_name['Uint64MapEntry']
_TESTNESTEDMAP_STRINGMAPENTRY = _TESTNESTEDMAP.nested_types_by_name['StringMapEntry']
_TESTNESTEDMAP_MAPMAPENTRY = _TESTNESTEDMAP.nested_types_by_name['MapMapEntry']
_TESTSTRINGMAP = DESCRIPTOR.message_types_by_name['TestStringMap']
_TESTSTRINGMAP_STRINGMAPENTRY = _TESTSTRINGMAP.nested_types_by_name['StringMapEntry']
_TESTWRAPPER = DESCRIPTOR.message_types_by_name['TestWrapper']
_TESTTIMESTAMP = DESCRIPTOR.message_types_by_name['TestTimestamp']
_TESTDURATION = DESCRIPTOR.message_types_by_name['TestDuration']
_TESTFIELDMASK = DESCRIPTOR.message_types_by_name['TestFieldMask']
_TESTSTRUCT = DESCRIPTOR.message_types_by_name['TestStruct']
_TESTANY = DESCRIPTOR.message_types_by_name['TestAny']
_TESTVALUE = DESCRIPTOR.message_types_by_name['TestValue']
_TESTLISTVALUE = DESCRIPTOR.message_types_by_name['TestListValue']
_TESTBOOLVALUE = DESCRIPTOR.message_types_by_name['TestBoolValue']
_TESTBOOLVALUE_BOOLMAPENTRY = _TESTBOOLVALUE.nested_types_by_name['BoolMapEntry']
_TESTCUSTOMJSONNAME = DESCRIPTOR.message_types_by_name['TestCustomJsonName']
_TESTEXTENSIONS = DESCRIPTOR.message_types_by_name['TestExtensions']
_TESTENUMVALUE = DESCRIPTOR.message_types_by_name['TestEnumValue']
MessageType = _reflection.GeneratedProtocolMessageType('MessageType', (_message.Message,), {
  'DESCRIPTOR' : _MESSAGETYPE,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.MessageType)
  })
_sym_db.RegisterMessage(MessageType)

TestMessage = _reflection.GeneratedProtocolMessageType('TestMessage', (_message.Message,), {
  'DESCRIPTOR' : _TESTMESSAGE,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestMessage)
  })
_sym_db.RegisterMessage(TestMessage)

TestOneof = _reflection.GeneratedProtocolMessageType('TestOneof', (_message.Message,), {
  'DESCRIPTOR' : _TESTONEOF,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestOneof)
  })
_sym_db.RegisterMessage(TestOneof)

TestMap = _reflection.GeneratedProtocolMessageType('TestMap', (_message.Message,), {

  'BoolMapEntry' : _reflection.GeneratedProtocolMessageType('BoolMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTMAP_BOOLMAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestMap.BoolMapEntry)
    })
  ,

  'Int32MapEntry' : _reflection.GeneratedProtocolMessageType('Int32MapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTMAP_INT32MAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestMap.Int32MapEntry)
    })
  ,

  'Int64MapEntry' : _reflection.GeneratedProtocolMessageType('Int64MapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTMAP_INT64MAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestMap.Int64MapEntry)
    })
  ,

  'Uint32MapEntry' : _reflection.GeneratedProtocolMessageType('Uint32MapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTMAP_UINT32MAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestMap.Uint32MapEntry)
    })
  ,

  'Uint64MapEntry' : _reflection.GeneratedProtocolMessageType('Uint64MapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTMAP_UINT64MAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestMap.Uint64MapEntry)
    })
  ,

  'StringMapEntry' : _reflection.GeneratedProtocolMessageType('StringMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTMAP_STRINGMAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestMap.StringMapEntry)
    })
  ,
  'DESCRIPTOR' : _TESTMAP,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestMap)
  })
_sym_db.RegisterMessage(TestMap)
_sym_db.RegisterMessage(TestMap.BoolMapEntry)
_sym_db.RegisterMessage(TestMap.Int32MapEntry)
_sym_db.RegisterMessage(TestMap.Int64MapEntry)
_sym_db.RegisterMessage(TestMap.Uint32MapEntry)
_sym_db.RegisterMessage(TestMap.Uint64MapEntry)
_sym_db.RegisterMessage(TestMap.StringMapEntry)

TestNestedMap = _reflection.GeneratedProtocolMessageType('TestNestedMap', (_message.Message,), {

  'BoolMapEntry' : _reflection.GeneratedProtocolMessageType('BoolMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTNESTEDMAP_BOOLMAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestNestedMap.BoolMapEntry)
    })
  ,

  'Int32MapEntry' : _reflection.GeneratedProtocolMessageType('Int32MapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTNESTEDMAP_INT32MAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestNestedMap.Int32MapEntry)
    })
  ,

  'Int64MapEntry' : _reflection.GeneratedProtocolMessageType('Int64MapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTNESTEDMAP_INT64MAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestNestedMap.Int64MapEntry)
    })
  ,

  'Uint32MapEntry' : _reflection.GeneratedProtocolMessageType('Uint32MapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTNESTEDMAP_UINT32MAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestNestedMap.Uint32MapEntry)
    })
  ,

  'Uint64MapEntry' : _reflection.GeneratedProtocolMessageType('Uint64MapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTNESTEDMAP_UINT64MAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestNestedMap.Uint64MapEntry)
    })
  ,

  'StringMapEntry' : _reflection.GeneratedProtocolMessageType('StringMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTNESTEDMAP_STRINGMAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestNestedMap.StringMapEntry)
    })
  ,

  'MapMapEntry' : _reflection.GeneratedProtocolMessageType('MapMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTNESTEDMAP_MAPMAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestNestedMap.MapMapEntry)
    })
  ,
  'DESCRIPTOR' : _TESTNESTEDMAP,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestNestedMap)
  })
_sym_db.RegisterMessage(TestNestedMap)
_sym_db.RegisterMessage(TestNestedMap.BoolMapEntry)
_sym_db.RegisterMessage(TestNestedMap.Int32MapEntry)
_sym_db.RegisterMessage(TestNestedMap.Int64MapEntry)
_sym_db.RegisterMessage(TestNestedMap.Uint32MapEntry)
_sym_db.RegisterMessage(TestNestedMap.Uint64MapEntry)
_sym_db.RegisterMessage(TestNestedMap.StringMapEntry)
_sym_db.RegisterMessage(TestNestedMap.MapMapEntry)

TestStringMap = _reflection.GeneratedProtocolMessageType('TestStringMap', (_message.Message,), {

  'StringMapEntry' : _reflection.GeneratedProtocolMessageType('StringMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTSTRINGMAP_STRINGMAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestStringMap.StringMapEntry)
    })
  ,
  'DESCRIPTOR' : _TESTSTRINGMAP,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestStringMap)
  })
_sym_db.RegisterMessage(TestStringMap)
_sym_db.RegisterMessage(TestStringMap.StringMapEntry)

TestWrapper = _reflection.GeneratedProtocolMessageType('TestWrapper', (_message.Message,), {
  'DESCRIPTOR' : _TESTWRAPPER,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestWrapper)
  })
_sym_db.RegisterMessage(TestWrapper)

TestTimestamp = _reflection.GeneratedProtocolMessageType('TestTimestamp', (_message.Message,), {
  'DESCRIPTOR' : _TESTTIMESTAMP,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestTimestamp)
  })
_sym_db.RegisterMessage(TestTimestamp)

TestDuration = _reflection.GeneratedProtocolMessageType('TestDuration', (_message.Message,), {
  'DESCRIPTOR' : _TESTDURATION,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestDuration)
  })
_sym_db.RegisterMessage(TestDuration)

TestFieldMask = _reflection.GeneratedProtocolMessageType('TestFieldMask', (_message.Message,), {
  'DESCRIPTOR' : _TESTFIELDMASK,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestFieldMask)
  })
_sym_db.RegisterMessage(TestFieldMask)

TestStruct = _reflection.GeneratedProtocolMessageType('TestStruct', (_message.Message,), {
  'DESCRIPTOR' : _TESTSTRUCT,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestStruct)
  })
_sym_db.RegisterMessage(TestStruct)

TestAny = _reflection.GeneratedProtocolMessageType('TestAny', (_message.Message,), {
  'DESCRIPTOR' : _TESTANY,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestAny)
  })
_sym_db.RegisterMessage(TestAny)

TestValue = _reflection.GeneratedProtocolMessageType('TestValue', (_message.Message,), {
  'DESCRIPTOR' : _TESTVALUE,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestValue)
  })
_sym_db.RegisterMessage(TestValue)

TestListValue = _reflection.GeneratedProtocolMessageType('TestListValue', (_message.Message,), {
  'DESCRIPTOR' : _TESTLISTVALUE,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestListValue)
  })
_sym_db.RegisterMessage(TestListValue)

TestBoolValue = _reflection.GeneratedProtocolMessageType('TestBoolValue', (_message.Message,), {

  'BoolMapEntry' : _reflection.GeneratedProtocolMessageType('BoolMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _TESTBOOLVALUE_BOOLMAPENTRY,
    '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
    # @@protoc_insertion_point(class_scope:proto3.TestBoolValue.BoolMapEntry)
    })
  ,
  'DESCRIPTOR' : _TESTBOOLVALUE,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestBoolValue)
  })
_sym_db.RegisterMessage(TestBoolValue)
_sym_db.RegisterMessage(TestBoolValue.BoolMapEntry)

TestCustomJsonName = _reflection.GeneratedProtocolMessageType('TestCustomJsonName', (_message.Message,), {
  'DESCRIPTOR' : _TESTCUSTOMJSONNAME,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestCustomJsonName)
  })
_sym_db.RegisterMessage(TestCustomJsonName)

TestExtensions = _reflection.GeneratedProtocolMessageType('TestExtensions', (_message.Message,), {
  'DESCRIPTOR' : _TESTEXTENSIONS,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestExtensions)
  })
_sym_db.RegisterMessage(TestExtensions)

TestEnumValue = _reflection.GeneratedProtocolMessageType('TestEnumValue', (_message.Message,), {
  'DESCRIPTOR' : _TESTENUMVALUE,
  '__module__' : 'google.protobuf.util.json_format_proto3_pb2'
  # @@protoc_insertion_point(class_scope:proto3.TestEnumValue)
  })
_sym_db.RegisterMessage(TestEnumValue)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030com.google.protobuf.utilB\020JsonFormatProto3'
  _TESTMAP_BOOLMAPENTRY._options = None
  _TESTMAP_BOOLMAPENTRY._serialized_options = b'8\001'
  _TESTMAP_INT32MAPENTRY._options = None
  _TESTMAP_INT32MAPENTRY._serialized_options = b'8\001'
  _TESTMAP_INT64MAPENTRY._options = None
  _TESTMAP_INT64MAPENTRY._serialized_options = b'8\001'
  _TESTMAP_UINT32MAPENTRY._options = None
  _TESTMAP_UINT32MAPENTRY._serialized_options = b'8\001'
  _TESTMAP_UINT64MAPENTRY._options = None
  _TESTMAP_UINT64MAPENTRY._serialized_options = b'8\001'
  _TESTMAP_STRINGMAPENTRY._options = None
  _TESTMAP_STRINGMAPENTRY._serialized_options = b'8\001'
  _TESTNESTEDMAP_BOOLMAPENTRY._options = None
  _TESTNESTEDMAP_BOOLMAPENTRY._serialized_options = b'8\001'
  _TESTNESTEDMAP_INT32MAPENTRY._options = None
  _TESTNESTEDMAP_INT32MAPENTRY._serialized_options = b'8\001'
  _TESTNESTEDMAP_INT64MAPENTRY._options = None
  _TESTNESTEDMAP_INT64MAPENTRY._serialized_options = b'8\001'
  _TESTNESTEDMAP_UINT32MAPENTRY._options = None
  _TESTNESTEDMAP_UINT32MAPENTRY._serialized_options = b'8\001'
  _TESTNESTEDMAP_UINT64MAPENTRY._options = None
  _TESTNESTEDMAP_UINT64MAPENTRY._serialized_options = b'8\001'
  _TESTNESTEDMAP_STRINGMAPENTRY._options = None
  _TESTNESTEDMAP_STRINGMAPENTRY._serialized_options = b'8\001'
  _TESTNESTEDMAP_MAPMAPENTRY._options = None
  _TESTNESTEDMAP_MAPMAPENTRY._serialized_options = b'8\001'
  _TESTSTRINGMAP_STRINGMAPENTRY._options = None
  _TESTSTRINGMAP_STRINGMAPENTRY._serialized_options = b'8\001'
  _TESTBOOLVALUE_BOOLMAPENTRY._options = None
  _TESTBOOLVALUE_BOOLMAPENTRY._serialized_options = b'8\001'
  _ENUMTYPE._serialized_start=4849
  _ENUMTYPE._serialized_end=4877
  _MESSAGETYPE._serialized_start=277
  _MESSAGETYPE._serialized_end=305
  _TESTMESSAGE._serialized_start=308
  _TESTMESSAGE._serialized_end=968
  _TESTONEOF._serialized_start=971
  _TESTONEOF._serialized_end=1239
  _TESTMAP._serialized_start=1242
  _TESTMAP._serialized_end=1851
  _TESTMAP_BOOLMAPENTRY._serialized_start=1557
  _TESTMAP_BOOLMAPENTRY._serialized_end=1603
  _TESTMAP_INT32MAPENTRY._serialized_start=1605
  _TESTMAP_INT32MAPENTRY._serialized_end=1652
  _TESTMAP_INT64MAPENTRY._serialized_start=1654
  _TESTMAP_INT64MAPENTRY._serialized_end=1701
  _TESTMAP_UINT32MAPENTRY._serialized_start=1703
  _TESTMAP_UINT32MAPENTRY._serialized_end=1751
  _TESTMAP_UINT64MAPENTRY._serialized_start=1753
  _TESTMAP_UINT64MAPENTRY._serialized_end=1801
  _TESTMAP_STRINGMAPENTRY._serialized_start=1803
  _TESTMAP_STRINGMAPENTRY._serialized_end=1851
  _TESTNESTEDMAP._serialized_start=1854
  _TESTNESTEDMAP._serialized_end=2627
  _TESTNESTEDMAP_BOOLMAPENTRY._serialized_start=1557
  _TESTNESTEDMAP_BOOLMAPENTRY._serialized_end=1603
  _TESTNESTEDMAP_INT32MAPENTRY._serialized_start=1605
  _TESTNESTEDMAP_INT32MAPENTRY._serialized_end=1652
  _TESTNESTEDMAP_INT64MAPENTRY._serialized_start=1654
  _TESTNESTEDMAP_INT64MAPENTRY._serialized_end=1701
  _TESTNESTEDMAP_UINT32MAPENTRY._serialized_start=1703
  _TESTNESTEDMAP_UINT32MAPENTRY._serialized_end=1751
  _TESTNESTEDMAP_UINT64MAPENTRY._serialized_start=1753
  _TESTNESTEDMAP_UINT64MAPENTRY._serialized_end=1801
  _TESTNESTEDMAP_STRINGMAPENTRY._serialized_start=1803
  _TESTNESTEDMAP_STRINGMAPENTRY._serialized_end=1851
  _TESTNESTEDMAP_MAPMAPENTRY._serialized_start=2559
  _TESTNESTEDMAP_MAPMAPENTRY._serialized_end=2627
  _TESTSTRINGMAP._serialized_start=2629
  _TESTSTRINGMAP._serialized_end=2752
  _TESTSTRINGMAP_STRINGMAPENTRY._serialized_start=2704
  _TESTSTRINGMAP_STRINGMAPENTRY._serialized_end=2752
  _TESTWRAPPER._serialized_start=2755
  _TESTWRAPPER._serialized_end=3761
  _TESTTIMESTAMP._serialized_start=3763
  _TESTTIMESTAMP._serialized_end=3873
  _TESTDURATION._serialized_start=3875
  _TESTDURATION._serialized_end=3982
  _TESTFIELDMASK._serialized_start=3984
  _TESTFIELDMASK._serialized_end=4042
  _TESTSTRUCT._serialized_start=4044
  _TESTSTRUCT._serialized_end=4145
  _TESTANY._serialized_start=4147
  _TESTANY._serialized_end=4239
  _TESTVALUE._serialized_start=4241
  _TESTVALUE._serialized_end=4339
  _TESTLISTVALUE._serialized_start=4341
  _TESTLISTVALUE._serialized_end=4451
  _TESTBOOLVALUE._serialized_start=4454
  _TESTBOOLVALUE._serialized_end=4591
  _TESTBOOLVALUE_BOOLMAPENTRY._serialized_start=1557
  _TESTBOOLVALUE_BOOLMAPENTRY._serialized_end=1603
  _TESTCUSTOMJSONNAME._serialized_start=4593
  _TESTCUSTOMJSONNAME._serialized_end=4636
  _TESTEXTENSIONS._serialized_start=4638
  _TESTEXTENSIONS._serialized_end=4712
  _TESTENUMVALUE._serialized_start=4715
  _TESTENUMVALUE._serialized_end=4847
# @@protoc_insertion_point(module_scope)



"""
A resurrection of some old functions from Python 2 for use in Python 3. These
should be used sparingly, to help with porting efforts, since code using them
is no longer standard Python 3 code.

This module provides the following:

1. Implementations of these builtin functions which have no equivalent on Py3:

- apply
- chr
- cmp
- execfile

2. Aliases:

- intern <- sys.intern
- raw_input <- input
- reduce <- functools.reduce
- reload <- imp.reload
- unichr <- chr
- unicode <- str
- xrange <- range

3. List-producing versions of the corresponding Python 3 iterator-producing functions:

- filter
- map
- range
- zip

4. Forward-ported Py2 types:

- basestring
- dict
- str
- long
- unicode

"""

from ...future.utils import PY3
from ...past.builtins.noniterators import (filter, map, range, reduce, zip)
# from past.builtins.misc import (ascii, hex, input, oct, open)
if PY3:
    from ...past.types import (basestring,
                            olddict as dict,
                            oldstr as str,
                            long,
                            unicode)
else:
    from __builtin__ import (basestring, dict, str, long, unicode)

from ...past.builtins.misc import (apply, chr, cmp, execfile, intern, oct,
                                raw_input, reload, unichr, unicode, xrange)
from ...past import utils


if utils.PY3:
    # We only import names that shadow the builtins on Py3. No other namespace
    # pollution on Py3.

    # Only shadow builtins on Py3; no new names
    __all__ = ['filter', 'map', 'range', 'reduce', 'zip',
               'basestring', 'dict', 'str', 'long', 'unicode',
               'apply', 'chr', 'cmp', 'execfile', 'intern', 'raw_input',
               'reload', 'unichr', 'xrange'
              ]

else:
    # No namespace pollution on Py2
    __all__ = []



# Copyright 2015 Lukas Lalinsky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import collections

from lxdbapi.types import javaTypetoNative, nativeToParamType, datetimeToTimezoneUtc, timezoneFromDatetimeUtc, \
    LXARRAY, typed_value_2_native
from lxdbapi.errors import ProgrammingError, ArrayTypeException
from lxdbapi.common_pb2 import Rep, TypedValue
import datetime

__all__ = ['Cursor', 'ColumnDescription', 'DictCursor']

logger = logging.getLogger(__name__)

# TODO see note in Cursor.rowcount()
MAX_INT = 2 ** 64 - 1

ColumnDescription = collections.namedtuple('ColumnDescription', 'name type_code display_size internal_size precision scale null_ok')
"""Named tuple for representing results from :attr:`Cursor.description`."""


class Cursor(object):
    """Database cursor for executing queries and iterating over results.

    You should not construct this object manually, use :meth:`Connection.cursor() <phoenixdb.connection.Connection.cursor>` instead.
    """

    arraysize = 1
    """
    Read/write attribute specifying the number of rows to fetch
    at a time with :meth:`fetchmany`. It defaults to 1 meaning to
    fetch a single row at a time.
    """

    itersize = 2000
    """
    Read/write attribute specifying the number of rows to fetch
    from the backend at each network roundtrip during iteration
    on the cursor. The default is 2000.
    """

    def __init__(self, connection):
        self._connection = connection
        self._xid = None
        self._id = None
        self._old_id = None
        self._prepared_sql = None
        self._signature = None
        self._column_data_types = []
        self._frame = None
        self._pos = None
        self._closed = False
        self.arraysize = self.__class__.arraysize
        self.itersize = self.__class__.itersize
        self._updatecount = -1
        self.lastrowid = None

    def __del__(self):
        if not self._connection.closed and not self._closed:
            try:
                self.close()
            except BaseException as ex:
                logger.debug("Ignoring exception while closing: {}".format(ex))
            finally:
                self._closed = True
                logger.debug("del: Closed cursor {}".format(self._xid))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if not self._closed:
            try:
                self.close()
            except BaseException as ex:
                logger.debug("Ignoring exception while closing: {}".format(ex))
            finally:
                self._closed = True
                logger.debug("exit: Closed cursor {}".format(self._xid))

    def __iter__(self):
        return self

    def __next__(self):
        row = self.fetchone()
        if row is None:
            raise StopIteration
        return row

    next = __next__

    def close(self):
        """Closes the cursor.
        No further operations are allowed once the cursor is closed.

        If the cursor is used in a ``with`` statement, this method will
        be automatically called at the end of the ``with`` block.
        """
        if self._closed:
            return
        if self._id is not None:
            try:
                self._connection._client.close_statement(self._connection._id, self._id)
            except BaseException as ex:
                logger.debug('Ignoring exception: {}'.format(ex))
            self._old_id = self._id
            self._id = None
            self._prepared_sql = None
        self._signature = None
        self._column_data_types = []
        self._frame = None
        self._pos = None
        self._closed = True

    @property
    def closed(self):
        """Read-only attribute specifying if the cursor is closed or not."""
        return self._closed

    @property
    def description(self):
        if self._signature is None:
            return None
        description = []
        for column in self._signature.columns:
            description.append(ColumnDescription(
                column.column_name,
                column.type.name,
                column.display_size,
                None,
                column.precision,
                column.scale,
                None if column.nullable == 2 else bool(column.nullable),
            ))
        return description

    def set_xid(self, xid):
        self._xid = xid
        logger.debug('new xid [{}-{}]'. format(self._xid, self._id))

    def _set_id(self, id):
        if self._id is not None and self._id != id:
            logger.debug('[{}]Closing statement {}'.format(self._xid, self._id))
            self._prepared_sql = None
            self._connection._client.close_statement(self._connection._id, self._id)
        logger.debug('[{}]Setting statement id {}'.format(self._xid, id))
        self._old_id = self._id
        self._id = id

    def _set_type(self, types, attribute_string, return_list):
        for column in types:
            return_list.append(javaTypetoNative(getattr(column, attribute_string)))

    def _set_signature(self, signature):
        self._signature = signature
        self._column_data_types = []
        self._parameter_data_types = []
        if signature is None:
            return

        self._set_type(signature.columns, 'column_class_name', self._column_data_types)

        self._set_type(signature.parameters, 'class_name', self._parameter_data_types)


    def _set_frame(self, frame):
        self._log_frame("set frame(previous): ")
        self._frame = frame
        self._pos = None

        if frame is not None:
            if frame.rows:
                self._pos = 0
            #elif not frame.done:
                #print('got an empty frame, but the statement is not done yet')
                #raise InternalError('got an empty frame, but the statement is not done yet')
        self._log_frame("set frame: ")

    def _log_frame(self, prefix):
        if logger.isEnabledFor(logging.DEBUG):
            if self._frame:
                if self._frame.rows:
                    nlines = str(len(self._frame.rows))
                else:
                    nlines = 'frame.rows is null'
            else:
                nlines = 'frame is null'
            logger.debug('{} statement [{}-{}], pos [{}], frame length [{}]'.
                         format(prefix, self._xid, self._id, self._pos, nlines))

    def _fetch_next_frame(self):
        offset = self._frame.offset + len(self._frame.rows)
        logger.debug('next frame: statement [{}-{}], pos [{}], offset [{}]'.
                     format(self._xid, self._id, self._pos, offset))
        frame = self._connection._client.fetch(
            self._connection._id, self._id,
            offset=offset, frame_max_size=self.itersize)
        self._set_frame(frame)

    def _process_results(self, results):
        if results:
            result = results[0]
            if result.own_statement:
                logger.debug('Own statement: {}'.format(result.statement_id))
                self._set_id(result.statement_id)
            logger.debug('first frame for statement [{}]'.format(self._id))
            self._set_signature(result.signature)
            self._set_frame(result.first_frame)
            self._updatecount = result.update_count
        else:
            logger.debug('no results for cursor [{}]'.format(self._xid))


    def _transform_parameters(self, parameters):
        self._parameter_data_types = []
        parameters_list = list(parameters)
        for idx, param in enumerate(parameters_list):
            if isinstance(param, datetime.datetime):  # only for datetimes
                param = datetimeToTimezoneUtc(param, self._connection.useTZ)
                parameters_list[idx] = param
            typeParam = nativeToParamType(param)
            self._parameter_data_types.append(typeParam)

        typed_parameters = []
        for value, data_type in zip(parameters_list, self._parameter_data_types):

            if value is None:
                typed_parameters.append(TypedValue(null=True, type=Rep.Value('NULL')))
            elif isinstance(value, LXARRAY):

                v = TypedValue()  # parent typed value
                v.type = Rep.Value(data_type[0])
                if value.gettype() is int:
                    v.component_type = Rep.Value("LONG")
                elif value.gettype() is float:
                    v.component_type = Rep.Value("DOUBLE")
                elif value.gettype() is str:
                    v.component_type = Rep.Value("STRING")
                else:
                    raise ArrayTypeException("Array cannot be of type: {}".format(value.gettype()))

                for item in value.getvalue():
                    typeParam = nativeToParamType(item, isArray=True)  # int bigint varchar ???
                    v_temp = TypedValue()
                    v_temp.type = Rep.Value(typeParam[0])
                    if item != None:
                        setattr(v_temp, typeParam[2], typeParam[1](item))
                    else:
                        setattr(v_temp, typeParam[2], 1)  # 1 for True
                    v.array_value.append(v_temp)
                typed_parameters.append(v)
            else:
                v = TypedValue()
                v.type = Rep.Value(data_type[0])
                if data_type[1] is not None:
                    setattr(v, data_type[2], data_type[1](value))
                else:
                    setattr(v, data_type[2], value)

                typed_parameters.append(v)
        return typed_parameters

    def get_tables(self, catalog=None, schemaPattern=None, tableNamePattern=None, typeList=None):
        if self._closed:
            raise ProgrammingError('the cursor is already closed(get_tables)')
        results = self._connection._client.get_tables(self._connection._id, catalog,
                                                      schemaPattern, tableNamePattern, typeList)

        self._process_results([results])

    def get_schemas(self, catalog=None, schemaPattern=None):
        if self._closed:
            raise ProgrammingError('the cursor is already closed(get_schemas)')
        results = self._connection._client.get_schemas(
            self._connection._id, catalog, schemaPattern)

        self._process_results([results])

    def get_catalogs(self):
        if self._closed:
            raise ProgrammingError('the cursor is already closed(get_catalogs)')
        results = self._connection._client.get_catalogs(
            self._connection._id)

        self._process_results([results])

    def get_columns(self, catalog=None, schemaPattern=None, tableNamePattern=None, columnNamePattern=None):
        if self._closed:
            raise ProgrammingError('the cursor is already closed(get_columns)')

        results = self._connection._client.get_columns(self._connection._id, catalog,
                                                      schemaPattern, tableNamePattern, columnNamePattern)
        self._process_results([results])

    def get_table_types(self):
        if self._closed:
            raise ProgrammingError('the cursor is already closed(get_table_types)')
        results = self._connection._client.get_table_types(
            self._connection._id)

        self._process_results([results])

    def get_type_info(self):
        if self._closed:
            raise ProgrammingError('the cursor is already closed(get_type_info)')
        results = self._connection._client.get_type_info(
            self._connection._id)

        self._process_results([results])

    def execute(self, operation, parameters=None):
        if self._closed:
            raise ProgrammingError('the cursor is already closed(execute)')
        self._updatecount = -1
        self._set_frame(None)
        if parameters is None or len(parameters) == 0:
            if self._id is None:
                logger.debug('(execute)Creating statement')
                self._set_id(self._connection._client.create_statement(self._connection._id))
            logger.debug('(execute)Prepare and Execute statement {}: {}'.format(self._id, operation))
            results = self._connection._client.prepare_and_execute(
                self._connection._id, self._id,
                operation, -1, self.itersize)
            self._process_results(results)
        else:
            if operation is not self._prepared_sql:
                statement = self._connection._client.prepare(self._connection._id, operation, -1)
                logger.debug('(execute)Prepared statement {}: {}'.format(statement.id, operation))
                self._set_id(statement.id)
                self._set_signature(statement.signature)
                self._prepared_sql = operation

            logger.debug('(execute)Executing statement {}'.format(self._id))
            results = self._connection._client.execute(
                self._connection._id, self._id, self._signature, self._transform_parameters(parameters),
                first_frame_max_size=self.itersize)

            self._process_results(results)

    def executemany(self, operation, seq_of_parameters):
        if self._closed:
            raise ProgrammingError('the cursor is already closed(executemany)')
        op = operation.split()[0]
        if op.upper() == "SELECT":
            raise ProgrammingError('You cannot execute SELECT statements in executemany()')
        self._updatecount = -1
        self._set_frame(None)
        if operation is not self._prepared_sql:
            statement = self._connection._client.prepare(self._connection._id, operation, max_rows_total=-1)
            logger.debug('(executemany)Prepared statement id {}: {}'.format(statement.id, operation))
            self._set_id(statement.id)
            self._set_signature(statement.signature)
            self._prepared_sql = operation
        logger.debug('(executemany)Executing statement id {}'.format(self._id))
        update_counts = self._connection._client.executeBatch(self._connection._id, self._id, self._signature,
                                         [self._transform_parameters(parameters) for parameters in seq_of_parameters], first_frame_max_size=self.itersize)
        return update_counts

    def executemany2(self, operation, seq_of_parameters):
        if self._closed:
            raise ProgrammingError('the cursor is already closed(executemany2)')
        self._updatecount = -1
        self._set_frame(None)
        statement = self._connection._client.prepare(
            self._connection._id, operation, 0)
        stmt_id = statement.get('id')
        logger.debug('(executemany2) Prepared statement {}: {}'.format(stmt_id, operation))
        self._set_id(stmt_id)
        self._set_signature(statement.get('signature'))
        for parameters in seq_of_parameters:
            logger.debug('(executemany2) Executing statement: {}'.format(stmt_id))
            self._connection._client.execute(
                statement, self.itersize,
                [parameters])


        '''self._connection._client.execute(
           self._connection._id, self._id,
           statement.signature, self._transform_parameters(parameters),
           first_frame_max_size=0)'''

    def _transform_row(self, row):
        """Transforms a Row into Python values.

        :param row:
            A ``common_pb2.Row`` object.

        :returns:
            A list of values casted into the correct Python types.

        :raises:
            NotImplementedError
        """
        tmp_row = []

        for i, column in enumerate(row.value):
            if column.has_array_value:
                raise NotImplementedError('array types are not supported')
            elif column.scalar_value.null:
                tmp_row.append(None)
            else:
                field_name, rep, mutate_to, cast_from = self._column_data_types[i]

                # get the value from the field_name
                value = getattr(column.scalar_value, field_name)

                # cast the value
                if cast_from is not None:
                    value = cast_from(value)

                tmp_row.append(value)
        return tmp_row

    def fetchone(self):
        result_row = []
        self._log_frame("fetchone: ")
        if self._frame is None:
            logger.debug('fetchone from null frame: id [{}-{}], previous id [{}]'.
                         format(self._xid, self._id, self._old_id))
            return None
        if self._pos is None:
            return None
        rows = self._frame.rows
        row = rows[self._pos]
        self._pos += 1
        if self._pos >= len(rows):
            self._pos = None
            if not self._frame.done:
                self._fetch_next_frame()
        for value, data_type in zip(row.value, self._column_data_types):
            if value.scalar_value.null:
                result_row.append(None)
            else:
                if data_type[1] is not None:
                    val = data_type[1](getattr(value.scalar_value, data_type[2]))
                    if isinstance(val, datetime.datetime):
                        val = timezoneFromDatetimeUtc(val, self._connection.useTZ)
                    result_row.append(val)
                else:
                    result_row.append(typed_value_2_native(value.scalar_value))
                    # result_row.append(getattr(value.scalar_value, data_type[2]))

        return tuple(result_row)

    def fetchone2(self):
        self._log_frame("fetchone2: ")
        if self._frame is None:
            logger.debug('fetchone2 from null frame: id [{}-{}], previous id [{}]'.
                         format(self._xid, self._id, self._old_id))
            return None
        if self._pos is None:
            return None
        rows = self._frame.rows
        row = rows[self._pos]
        #row = self._transform_row(rows[self._pos])
        self._pos += 1
        if self._pos >= len(rows):
            self._pos = None
            if not self._frame.done:
                self._fetch_next_frame()
        return [row]

    def fetchmany(self, size=None):
        if size is None:
            size = self.arraysize
        rows = []
        while size > 0:
            row = self.fetchone()
            if row is None:
                break
            rows.append(row)
            size -= 1
        return rows

    def fetchall(self):
        rows = []
        while True:
            row = self.fetchone()
            if row is None:
                break
            rows.append(row)
        return rows

    def setinputsizes(self, sizes):
        pass

    def setoutputsize(self, size, column=None):
        pass

    @property
    def connection(self):
        """Read-only attribute providing access to the :class:`Connection <phoenixdb.connection.Connection>`
        object this cursor was created from."""
        return self._connection

    @property
    def rowcount(self):
        """Read-only attribute specifying the number of rows affected by
        the last executed DML statement or -1 if the number cannot be
        determined. Note that this will always be set to -1 for select
        queries."""
        # TODO instead of -1, this ends up being set to Integer.MAX_VALUE
        if self._updatecount == MAX_INT:
            return -1
        return self._updatecount

    @property
    def rownumber(self):
        """Read-only attribute providing the current 0-based index of the
        cursor in the result set or ``None`` if the index cannot be
        determined.

        The index can be seen as index of the cursor in a sequence
        (the result set). The next fetch operation will fetch the
        row indexed by :attr:`rownumber` in that sequence.
        """
        if self._frame is not None and self._pos is not None:
            return self._frame.offset + self._pos
        return self._pos


class DictCursor(Cursor):
    """A cursor which returns results as a dictionary"""

    def _transform_row(self, row):
        row = super(DictCursor, self)._transform_row(row)
        d = {}
        for ind, val in enumerate(row):
            d[self._signature.columns[ind].column_name] = val
        return d



# Copyright 2015 Lukas Lalinsky
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import time
import datetime
from decimal import Decimal

__all__ = [
    'Date', 'Time', 'Timestamp', 'DateFromTicks', 'TimeFromTicks', 'TimestampFromTicks',
    'Binary', 'STRING', 'BINARY', 'NUMBER', 'DATETIME', 'ROWID', 'BOOLEAN',
    'JAVA_CLASSES', 'JAVA_CLASSES_MAP', 'TypeHelper',
]


def Date(year, month, day):
    """Constructs an object holding a date value."""
    return datetime.date(year, month, day)


def Time(hour, minute, second):
    """Constructs an object holding a time value."""
    return datetime.time(hour, minute, second)


def Timestamp(year, month, day, hour, minute, second):
    """Constructs an object holding a datetime/timestamp value."""
    return datetime.datetime(year, month, day, hour, minute, second)


def DateFromTicks(ticks):
    """Constructs an object holding a date value from the given UNIX timestamp."""
    return Date(*time.localtime(ticks)[:3])


def TimeFromTicks(ticks):
    """Constructs an object holding a time value from the given UNIX timestamp."""
    return Time(*time.localtime(ticks)[3:6])


def TimestampFromTicks(ticks):
    """Constructs an object holding a datetime/timestamp value from the given UNIX timestamp."""
    return Timestamp(*time.localtime(ticks)[:6])


def Binary(value):
    """Constructs an object capable of holding a binary (long) string value."""
    return bytes(value)


def time_from_java_sql_time(n):
    dt = datetime.datetime(1970, 1, 1) + datetime.timedelta(milliseconds=n)
    return dt.time()


def time_to_java_sql_time(t):
    return ((t.hour * 60 + t.minute) * 60 + t.second) * 1000 + t.microsecond // 1000


def date_from_java_sql_date(n):
    return datetime.date(1970, 1, 1) + datetime.timedelta(days=n)


def date_to_java_sql_date(d):
    if isinstance(d, datetime.datetime):
        d = d.date()
    td = d - datetime.date(1970, 1, 1)
    return td.days


def datetime_from_java_sql_timestamp(n):
    return datetime.datetime(1970, 1, 1) + datetime.timedelta(milliseconds=n)


def datetime_to_java_sql_timestamp(d):
    td = d - datetime.datetime(1970, 1, 1)
    return td.microseconds // 1000 + (td.seconds + td.days * 24 * 3600) * 1000


class ColumnType(object):

    def __init__(self, eq_types):
        self.eq_types = tuple(eq_types)
        self.eq_types_set = set(eq_types)

    def __eq__(self, other):
        return other in self.eq_types_set

    def __cmp__(self, other):
        if other in self.eq_types_set:
            return 0
        if other < self.eq_types:
            return 1
        else:
            return -1


STRING = ColumnType(['VARCHAR', 'CHAR'])
"""Type object that can be used to describe string-based columns."""

BINARY = ColumnType(['BINARY', 'VARBINARY'])
"""Type object that can be used to describe (long) binary columns."""

NUMBER = ColumnType([
    'INTEGER', 'UNSIGNED_INT', 'BIGINT', 'UNSIGNED_LONG', 'TINYINT', 'UNSIGNED_TINYINT',
    'SMALLINT', 'UNSIGNED_SMALLINT', 'FLOAT', 'UNSIGNED_FLOAT', 'DOUBLE', 'UNSIGNED_DOUBLE', 'DECIMAL'
])
"""Type object that can be used to describe numeric columns."""

DATETIME = ColumnType(['TIME', 'DATE', 'TIMESTAMP', 'UNSIGNED_TIME', 'UNSIGNED_DATE', 'UNSIGNED_TIMESTAMP'])
"""Type object that can be used to describe date/time columns."""

ROWID = ColumnType([])
"""Only implemented for DB API 2.0 compatibility, not used."""

BOOLEAN = ColumnType(['BOOLEAN'])
"""Type object that can be used to describe boolean columns. This is a phoenixdb-specific extension."""


# XXX ARRAY

if sys.version_info[0] < 3:
    _long = int  # noqa: F821  IN PYTHON 2 THIS SHOULD BE CHANGED TO LONG
else:
    _long = int

JAVA_CLASSES = {
    'bool_value': [
        ('java.lang.Boolean', None, None),
    ],
    'string_value': [
        ('java.lang.Character',  None, None),
        ('java.lang.String',  None, None),
        ('java.math.BigDecimal',  str, Decimal),
    ],
    'number_value': [
        ('java.lang.Integer',  None, int),
        ('java.lang.Short',  None, int),
        ('java.lang.Long',  None, _long),
        ('java.lang.Byte',  None, int),
        ('java.sql.Time',  time_to_java_sql_time, time_from_java_sql_time),
        ('java.sql.Date',  date_to_java_sql_date, date_from_java_sql_date),
        ('java.sql.Timestamp',  datetime_to_java_sql_timestamp, datetime_from_java_sql_timestamp),
    ],
    'bytes_value': [
        ('[B',  Binary, None),
    ],
    'double_value': [
        # if common_pb2.FLOAT is used, incorrect values are sent
        ('java.lang.Float',  float, float),
        ('java.lang.Double',  float, float),
    ], 'object_value': [
        ('java.lang.Object', None, None)
    ]
}
"""Groups of Java classes."""

JAVA_CLASSES_MAP = dict((v[0], (k, v[1], v[2])) for k in JAVA_CLASSES for v in JAVA_CLASSES[k])
"""Flips the available types to allow for faster lookup by Java class.

This mapping should be structured as:
    {
        'java.math.BigDecimal': ('string_value', common_pb2.BIG_DECIMAL, str, Decimal),),
        ...
        '<java class>': (<field_name>, <Rep enum>, <mutate_to function>, <cast_from function>),
    }
"""


class TypeHelper(object):
    @staticmethod
    def from_class(klass):
        """Retrieves a Rep and functions to cast to/from based on the Java class.

        :param klass:
            The string of the Java class for the column or parameter.

        :returns: tuple ``(field_name, rep, mutate_to, cast_from)``
            WHERE
            ``field_name`` is the attribute in ``common_pb2.TypedValue``
            ``rep`` is the common_pb2.Rep enum
            ``mutate_to`` is the function to cast values into Phoenix values, if any
            ``cast_from`` is the function to cast from the Phoenix value to the Python value, if any

        :raises:
            NotImplementedError
        """
        if klass not in JAVA_CLASSES_MAP:
            raise NotImplementedError('type {} is not supported'.format(klass))

        return JAVA_CLASSES_MAP[klass]



##############################################################################
# Copyright (c) 2020 Zope Foundation and Contributors.
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
##############################################################################
"""
Interface definitions paralleling the abstract base classes defined in
:mod:`collections.abc`.

After this module is imported, the standard library types will declare
that they implement the appropriate interface. While most standard
library types will properly implement that interface (that
is, ``verifyObject(ISequence, list()))`` will pass, for example), a few might not:

    - `memoryview` doesn't feature all the defined methods of
      ``ISequence`` such as ``count``; it is still declared to provide
      ``ISequence`` though.

    - `collections.deque.pop` doesn't accept the ``index`` argument of
      `collections.abc.MutableSequence.pop`

    - `range.index` does not accept the ``start`` and ``stop`` arguments.

.. versionadded:: 5.0.0
"""

import sys

from abc import ABCMeta
from collections import abc
from collections import OrderedDict
from collections import UserList
from collections import UserDict
from collections import UserString

from ....zope.interface.common import ABCInterface
from ....zope.interface.common import optional

# pylint:disable=inherit-non-class,
# pylint:disable=no-self-argument,no-method-argument
# pylint:disable=unexpected-special-method-signature
# pylint:disable=no-value-for-parameter


def _new_in_ver(name, ver,
                bases_if_missing=(ABCMeta,),
                register_if_missing=()):
    if ver:
        return getattr(abc, name)

    # TODO: It's a shame to have to repeat the bases when
    # the ABC is missing. Can we DRY that?
    missing = ABCMeta(name, bases_if_missing, {
        '__doc__': "The ABC %s is not defined in this version of Python." % (
            name
        ),
    })

    for c in register_if_missing:
        missing.register(c)

    return missing

__all__ = [
    'IAsyncGenerator',
    'IAsyncIterable',
    'IAsyncIterator',
    'IAwaitable',
    'ICollection',
    'IContainer',
    'ICoroutine',
    'IGenerator',
    'IHashable',
    'IItemsView',
    'IIterable',
    'IIterator',
    'IKeysView',
    'IMapping',
    'IMappingView',
    'IMutableMapping',
    'IMutableSequence',
    'IMutableSet',
    'IReversible',
    'ISequence',
    'ISet',
    'ISized',
    'IValuesView',
]

class IContainer(ABCInterface):
    abc = abc.Container

    @optional
    def __contains__(other):
        """
        Optional method. If not provided, the interpreter will use
        ``__iter__`` or the old ``__getitem__`` protocol
        to implement ``in``.
        """

class IHashable(ABCInterface):
    abc = abc.Hashable

class IIterable(ABCInterface):
    abc = abc.Iterable

    @optional
    def __iter__():
        """
        Optional method. If not provided, the interpreter will
        implement `iter` using the old ``__getitem__`` protocol.
        """

class IIterator(IIterable):
    abc = abc.Iterator

class IReversible(IIterable):
    abc = _new_in_ver('Reversible', True, (IIterable.getABC(),))

    @optional
    def __reversed__():
        """
        Optional method. If this isn't present, the interpreter
        will use ``__len__`` and ``__getitem__`` to implement the
        `reversed` builtin.
        """

class IGenerator(IIterator):
    # New in Python 3.5
    abc = _new_in_ver('Generator', True, (IIterator.getABC(),))


class ISized(ABCInterface):
    abc = abc.Sized


# ICallable is not defined because there's no standard signature.

class ICollection(ISized,
                  IIterable,
                  IContainer):
    abc = _new_in_ver('Collection', True,
                      (ISized.getABC(), IIterable.getABC(), IContainer.getABC()))


class ISequence(IReversible,
                ICollection):
    abc = abc.Sequence
    extra_classes = (UserString,)
    # On Python 2, basestring is registered as an ISequence, and
    # its subclass str is an IByteString. If we also register str as
    # an ISequence, that tends to lead to inconsistent resolution order.
    ignored_classes = (basestring,) if str is bytes else () # pylint:disable=undefined-variable

    @optional
    def __reversed__():
        """
        Optional method. If this isn't present, the interpreter
        will use ``__len__`` and ``__getitem__`` to implement the
        `reversed` builtin.
        """

    @optional
    def __iter__():
        """
        Optional method. If not provided, the interpreter will
        implement `iter` using the old ``__getitem__`` protocol.
        """

class IMutableSequence(ISequence):
    abc = abc.MutableSequence
    extra_classes = (UserList,)


class IByteString(ISequence):
    """
    This unifies `bytes` and `bytearray`.
    """
    abc = _new_in_ver('ByteString', True,
                      (ISequence.getABC(),),
                      (bytes, bytearray))


class ISet(ICollection):
    abc = abc.Set


class IMutableSet(ISet):
    abc = abc.MutableSet


class IMapping(ICollection):
    abc = abc.Mapping
    extra_classes = (dict,)
    # OrderedDict is a subclass of dict. On CPython 2,
    # it winds up registered as a IMutableMapping, which
    # produces an inconsistent IRO if we also try to register it
    # here.
    ignored_classes = (OrderedDict,)


class IMutableMapping(IMapping):
    abc = abc.MutableMapping
    extra_classes = (dict, UserDict,)
    ignored_classes = (OrderedDict,)

class IMappingView(ISized):
    abc = abc.MappingView


class IItemsView(IMappingView, ISet):
    abc = abc.ItemsView


class IKeysView(IMappingView, ISet):
    abc = abc.KeysView


class IValuesView(IMappingView, ICollection):
    abc = abc.ValuesView

    @optional
    def __contains__(other):
        """
        Optional method. If not provided, the interpreter will use
        ``__iter__`` or the old ``__len__`` and ``__getitem__`` protocol
        to implement ``in``.
        """

class IAwaitable(ABCInterface):
    abc = _new_in_ver('Awaitable', True)


class ICoroutine(IAwaitable):
    abc = _new_in_ver('Coroutine', True)


class IAsyncIterable(ABCInterface):
    abc = _new_in_ver('AsyncIterable', True)


class IAsyncIterator(IAsyncIterable):
    abc = _new_in_ver('AsyncIterator', True)


class IAsyncGenerator(IAsyncIterator):
    abc = _new_in_ver('AsyncGenerator', True)



u"""
Fixer for (metaclass=X) -> __metaclass__ = X
Some semantics (see PEP 3115) may be altered in the translation."""

from lib2to3 import fixer_base
from lib2to3.fixer_util import Name, syms, Node, Leaf, Newline, find_root
from lib2to3.pygram import token
from ...libfuturize.fixer_util import indentation, suitify
# from ..fixer_util import Name, syms, Node, Leaf, Newline, find_root, indentation, suitify

def has_metaclass(parent):
    results = None
    for node in parent.children:
        kids = node.children
        if node.type == syms.argument:
            if kids[0] == Leaf(token.NAME, u"metaclass") and \
                kids[1] == Leaf(token.EQUAL, u"=") and \
                kids[2]:
                #Hack to avoid "class X(=):" with this case.
                results = [node] + kids
                break
        elif node.type == syms.arglist:
            # Argument list... loop through it looking for:
            # Node(*, [*, Leaf(token.NAME, u"metaclass"), Leaf(token.EQUAL, u"="), Leaf(*, *)]
            for child in node.children:
                if results: break
                if child.type == token.COMMA:
                    #Store the last comma, which precedes the metaclass
                    comma = child
                elif type(child) == Node:
                    meta = equal = name = None
                    for arg in child.children:
                        if arg == Leaf(token.NAME, u"metaclass"):
                            #We have the (metaclass) part
                            meta = arg
                        elif meta and arg == Leaf(token.EQUAL, u"="):
                            #We have the (metaclass=) part
                            equal = arg
                        elif meta and equal:
                            #Here we go, we have (metaclass=X)
                            name = arg
                            results = (comma, meta, equal, name)
                            break
    return results


class FixMetaclass(fixer_base.BaseFix):

    PATTERN = u"""
    classdef<any*>
    """

    def transform(self, node, results):
        meta_results = has_metaclass(node)
        if not meta_results: return
        for meta in meta_results:
            meta.remove()
        target = Leaf(token.NAME, u"__metaclass__")
        equal = Leaf(token.EQUAL, u"=", prefix=u" ")
        # meta is the last item in what was returned by has_metaclass(): name
        name = meta
        name.prefix = u" "
        stmt_node = Node(syms.atom, [target, equal, name])

        suitify(node)
        for item in node.children:
            if item.type == syms.suite:
                for stmt in item.children:
                    if stmt.type == token.INDENT:
                        # Insert, in reverse order, the statement, a newline,
                        # and an indent right after the first indented line
                        loc = item.children.index(stmt) + 1
                        # Keep consistent indentation form
                        ident = Leaf(token.INDENT, stmt.value)
                        item.insert_child(loc, ident)
                        item.insert_child(loc, Newline())
                        item.insert_child(loc, stmt_node)
                        break
